from django.apps import apps
from django.conf import settings
from django.utils.module_loading import import_string


class AbstractThumbnailer(object):
    def generate_thumbnail(self, source, **opts):
        raise NotImplementedError

    def delete_thumbnails(self, source):
        raise NotImplementedError


class SorlThumbnail(AbstractThumbnailer):
    def __init__(self):
        if not apps.is_installed("sorl.thumbnail"):
            raise ValueError('"sorl.thumbnail" is not listed in "INSTALLED_APPS".')

    def generate_thumbnail(self, source, **opts):
        from sorl.thumbnail import get_thumbnail

        # Sorl can accept only: "width x height", "width", "x height".
        # https://sorl-thumbnail.readthedocs.io/en/latest/template.html#geometry
        # So for example value '50x' must be converted to '50'.
        size = opts.pop("size")
        width, height = size.split("x")
        # Set `size` to `width` if `height` is not provided.
        size = size if height else width
        return get_thumbnail(source, size, **opts)

    def delete_thumbnails(self, source):
        from sorl.thumbnail import delete
        from sorl.thumbnail.helpers import ThumbnailError

        try:
            delete(source)
        except ThumbnailError:
            pass


class EasyThumbnails(AbstractThumbnailer):
    def __init__(self):
        if not apps.is_installed("easy_thumbnails"):
            raise ValueError('"easy_thumbnails" is not listed in "INSTALLED_APPS".')

    def generate_thumbnail(self, source, **opts):
        from easy_thumbnails.files import get_thumbnailer as get_easy_thumbnailer

        width, height = opts["size"].split("x")
        width = width or 0
        height = height or 0
        opts["size"] = (width, height)
        return get_easy_thumbnailer(source).get_thumbnail(opts)

    def delete_thumbnails(self, source):
        from easy_thumbnails.files import get_thumbnailer as get_easy_thumbnailer

        get_easy_thumbnailer(source).delete(save=False)


def get_thumbnailer():
    thumbnailer = import_string(settings.OSCAR_THUMBNAILER)
    return thumbnailer()



import re
import zlib

from django.conf import settings
from django.core import exceptions
from django.db import models
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy
from phonenumber_field.modelfields import PhoneNumberField

from oscar.core.compat import AUTH_USER_MODEL
from oscar.models.fields import UppercaseCharField


class AbstractAddress(models.Model):
    """
    Superclass address object

    This is subclassed and extended to provide models for
    user, shipping and billing addresses.
    """

    MR, MISS, MRS, MS, DR = ("Mr", "Miss", "Mrs", "Ms", "Dr")
    TITLE_CHOICES = (
        (MR, _("Mr")),
        (MISS, _("Miss")),
        (MRS, _("Mrs")),
        (MS, _("Ms")),
        (DR, _("Dr")),
    )

    POSTCODE_REQUIRED = "postcode" in settings.OSCAR_REQUIRED_ADDRESS_FIELDS

    # Regex for each country. Not listed countries don't use postcodes
    # Based on http://en.wikipedia.org/wiki/List_of_postal_codes
    POSTCODES_REGEX = {
        "AC": r"^[A-Z]{4}[0-9][A-Z]$",
        "AD": r"^AD[0-9]{3}$",
        "AF": r"^[0-9]{4}$",
        "AI": r"^AI-2640$",
        "AL": r"^[0-9]{4}$",
        "AM": r"^[0-9]{4}$",
        "AR": r"^([0-9]{4}|[A-Z][0-9]{4}[A-Z]{3})$",
        "AS": r"^[0-9]{5}(-[0-9]{4}|-[0-9]{6})?$",
        "AT": r"^[0-9]{4}$",
        "AU": r"^[0-9]{4}$",
        "AX": r"^[0-9]{5}$",
        "AZ": r"^AZ[0-9]{4}$",
        "BA": r"^[0-9]{5}$",
        "BB": r"^BB[0-9]{5}$",
        "BD": r"^[0-9]{4}$",
        "BE": r"^[0-9]{4}$",
        "BG": r"^[0-9]{4}$",
        "BH": r"^[0-9]{3,4}$",
        "BL": r"^[0-9]{5}$",
        "BM": r"^[A-Z]{2}([0-9]{2}|[A-Z]{2})",
        "BN": r"^[A-Z]{2}[0-9]{4}$",
        "BO": r"^[0-9]{4}$",
        "BR": r"^[0-9]{5}(-[0-9]{3})?$",
        "BT": r"^[0-9]{3}$",
        "BY": r"^[0-9]{6}$",
        "CA": r"^[A-Z][0-9][A-Z][0-9][A-Z][0-9]$",
        "CC": r"^[0-9]{4}$",
        "CH": r"^[0-9]{4}$",
        "CL": r"^([0-9]{7}|[0-9]{3}-[0-9]{4})$",
        "CN": r"^[0-9]{6}$",
        "CO": r"^[0-9]{6}$",
        "CR": r"^[0-9]{4,5}$",
        "CU": r"^[0-9]{5}$",
        "CV": r"^[0-9]{4}$",
        "CX": r"^[0-9]{4}$",
        "CY": r"^[0-9]{4}$",
        "CZ": r"^[0-9]{5}$",
        "DE": r"^[0-9]{5}$",
        "DK": r"^[0-9]{4}$",
        "DO": r"^[0-9]{5}$",
        "DZ": r"^[0-9]{5}$",
        "EC": r"^EC[0-9]{6}$",
        "EE": r"^[0-9]{5}$",
        "EG": r"^[0-9]{5}$",
        "ES": r"^[0-9]{5}$",
        "ET": r"^[0-9]{4}$",
        "FI": r"^[0-9]{5}$",
        "FK": r"^[A-Z]{4}[0-9][A-Z]{2}$",
        "FM": r"^[0-9]{5}(-[0-9]{4})?$",
        "FO": r"^[0-9]{3}$",
        "FR": r"^[0-9]{5}$",
        "GA": r"^[0-9]{2}.*[0-9]{2}$",
        "GB": r"^[A-Z][A-Z0-9]{1,3}[0-9][A-Z]{2}$",
        "GE": r"^[0-9]{4}$",
        "GF": r"^[0-9]{5}$",
        "GG": r"^([A-Z]{2}[0-9]{2,3}[A-Z]{2})$",
        "GI": r"^GX111AA$",
        "GL": r"^[0-9]{4}$",
        "GP": r"^[0-9]{5}$",
        "GR": r"^[0-9]{5}$",
        "GS": r"^SIQQ1ZZ$",
        "GT": r"^[0-9]{5}$",
        "GU": r"^[0-9]{5}$",
        "GW": r"^[0-9]{4}$",
        "HM": r"^[0-9]{4}$",
        "HN": r"^[0-9]{5}$",
        "HR": r"^[0-9]{5}$",
        "HT": r"^[0-9]{4}$",
        "HU": r"^[0-9]{4}$",
        "ID": r"^[0-9]{5}$",
        "IL": r"^([0-9]{5}|[0-9]{7})$",
        "IM": r"^IM[0-9]{2,3}[A-Z]{2}$$",
        "IN": r"^[0-9]{6}$",
        "IO": r"^[A-Z]{4}[0-9][A-Z]{2}$",
        "IQ": r"^[0-9]{5}$",
        "IR": r"^[0-9]{5}-[0-9]{5}$",
        "IS": r"^[0-9]{3}$",
        "IT": r"^[0-9]{5}$",
        "JE": r"^JE[0-9]{2}[A-Z]{2}$",
        "JM": r"^JM[A-Z]{3}[0-9]{2}$",
        "JO": r"^[0-9]{5}$",
        "JP": r"^[0-9]{3}-?[0-9]{4}$",
        "KE": r"^[0-9]{5}$",
        "KG": r"^[0-9]{6}$",
        "KH": r"^[0-9]{5}$",
        "KR": r"^[0-9]{5}$",
        "KY": r"^KY[0-9]-[0-9]{4}$",
        "KZ": r"^[0-9]{6}$",
        "LA": r"^[0-9]{5}$",
        "LB": r"^[0-9]{8}$",
        "LI": r"^[0-9]{4}$",
        "LK": r"^[0-9]{5}$",
        "LR": r"^[0-9]{4}$",
        "LS": r"^[0-9]{3}$",
        "LT": r"^(LT-)?[0-9]{5}$",
        "LU": r"^[0-9]{4}$",
        "LV": r"^LV-[0-9]{4}$",
        "LY": r"^[0-9]{5}$",
        "MA": r"^[0-9]{5}$",
        "MC": r"^980[0-9]{2}$",
        "MD": r"^MD-?[0-9]{4}$",
        "ME": r"^[0-9]{5}$",
        "MF": r"^[0-9]{5}$",
        "MG": r"^[0-9]{3}$",
        "MH": r"^[0-9]{5}$",
        "MK": r"^[0-9]{4}$",
        "MM": r"^[0-9]{5}$",
        "MN": r"^[0-9]{5}$",
        "MP": r"^[0-9]{5}$",
        "MQ": r"^[0-9]{5}$",
        "MT": r"^[A-Z]{3}[0-9]{4}$",
        "MV": r"^[0-9]{4,5}$",
        "MX": r"^[0-9]{5}$",
        "MY": r"^[0-9]{5}$",
        "MZ": r"^[0-9]{4}$",
        "NA": r"^[0-9]{5}$",
        "NC": r"^[0-9]{5}$",
        "NE": r"^[0-9]{4}$",
        "NF": r"^[0-9]{4}$",
        "NG": r"^[0-9]{6}$",
        "NI": r"^[0-9]{5}$",
        "NL": r"^[0-9]{4}[A-Z]{2}$",
        "NO": r"^[0-9]{4}$",
        "NP": r"^[0-9]{5}$",
        "NZ": r"^[0-9]{4}$",
        "OM": r"^[0-9]{3}$",
        "PA": r"^[0-9]{6}$",
        "PE": r"^[0-9]{5}$",
        "PF": r"^[0-9]{5}$",
        "PG": r"^[0-9]{3}$",
        "PH": r"^[0-9]{4}$",
        "PK": r"^[0-9]{5}$",
        "PL": r"^[0-9]{2}-?[0-9]{3}$",
        "PM": r"^[0-9]{5}$",
        "PN": r"^[A-Z]{4}[0-9][A-Z]{2}$",
        "PR": r"^[0-9]{5}$",
        "PT": r"^[0-9]{4}(-?[0-9]{3})?$",
        "PW": r"^[0-9]{5}$",
        "PY": r"^[0-9]{4}$",
        "RE": r"^[0-9]{5}$",
        "RO": r"^[0-9]{6}$",
        "RS": r"^[0-9]{5}$",
        "RU": r"^[0-9]{6}$",
        "SA": r"^[0-9]{5}$",
        "SD": r"^[0-9]{5}$",
        "SE": r"^[0-9]{5}$",
        "SG": r"^([0-9]{2}|[0-9]{4}|[0-9]{6})$",
        "SH": r"^(STHL1ZZ|TDCU1ZZ)$",
        "SI": r"^(SI-)?[0-9]{4}$",
        "SK": r"^[0-9]{5}$",
        "SM": r"^[0-9]{5}$",
        "SN": r"^[0-9]{5}$",
        "SV": r"^01101$",
        "SZ": r"^[A-Z][0-9]{3}$",
        "TC": r"^TKCA1ZZ$",
        "TD": r"^[0-9]{5}$",
        "TH": r"^[0-9]{5}$",
        "TJ": r"^[0-9]{6}$",
        "TM": r"^[0-9]{6}$",
        "TN": r"^[0-9]{4}$",
        "TR": r"^[0-9]{5}$",
        "TT": r"^[0-9]{6}$",
        "TW": r"^([0-9]{3}|[0-9]{5})$",
        "UA": r"^[0-9]{5}$",
        "US": r"^[0-9]{5}(-[0-9]{4}|-[0-9]{6})?$",
        "UY": r"^[0-9]{5}$",
        "UZ": r"^[0-9]{6}$",
        "VA": r"^00120$",
        "VC": r"^VC[0-9]{4}",
        "VE": r"^[0-9]{4}[A-Z]?$",
        "VG": r"^VG[0-9]{4}$",
        "VI": r"^[0-9]{5}$",
        "VN": r"^[0-9]{6}$",
        "WF": r"^[0-9]{5}$",
        "XK": r"^[0-9]{5}$",
        "YT": r"^[0-9]{5}$",
        "ZA": r"^[0-9]{4}$",
        "ZM": r"^[0-9]{5}$",
    }

    title = models.CharField(
        pgettext_lazy("Treatment Pronouns for the customer", "Title"),
        max_length=64,
        choices=TITLE_CHOICES,
        blank=True,
    )
    first_name = models.CharField(_("First name"), max_length=255, blank=True)
    last_name = models.CharField(_("Last name"), max_length=255, blank=True)

    # We use quite a few lines of an address as they are often quite long and
    # it's easier to just hide the unnecessary ones than add extra ones.
    line1 = models.CharField(_("First line of address"), max_length=255)
    line2 = models.CharField(_("Second line of address"), max_length=255, blank=True)
    line3 = models.CharField(_("Third line of address"), max_length=255, blank=True)
    line4 = models.CharField(_("City"), max_length=255, blank=True)
    state = models.CharField(_("State/County"), max_length=255, blank=True)
    postcode = UppercaseCharField(_("Post/Zip-code"), max_length=64, blank=True)
    country = models.ForeignKey(
        "address.Country", on_delete=models.CASCADE, verbose_name=_("Country")
    )

    # A field only used for searching addresses - this contains all the
    # `search_fields`.  This is effectively a poor man's Solr text field.
    search_text = models.TextField(
        _("Search text - used only for searching addresses"), editable=False
    )
    search_fields = [
        "first_name",
        "last_name",
        "line1",
        "line2",
        "line3",
        "line4",
        "state",
        "postcode",
        "country",
    ]

    # Fields, used for `summary` property definition and hash generation.
    base_fields = hash_fields = [
        "salutation",
        "line1",
        "line2",
        "line3",
        "line4",
        "state",
        "postcode",
        "country",
    ]

    def __str__(self):
        return self.summary

    class Meta:
        abstract = True
        verbose_name = _("Address")
        verbose_name_plural = _("Addresses")

    # Saving

    def save(self, *args, **kwargs):
        self._update_search_text()
        super().save(*args, **kwargs)

    def clean(self):
        # Strip all whitespace
        for field in [
            "first_name",
            "last_name",
            "line1",
            "line2",
            "line3",
            "line4",
            "state",
            "postcode",
        ]:
            if self.__dict__[field]:
                self.__dict__[field] = self.__dict__[field].strip()

        # Ensure postcodes are valid for country
        self.ensure_postcode_is_valid_for_country()

    def ensure_postcode_is_valid_for_country(self):
        """
        Validate postcode given the country
        """
        if not self.postcode and self.POSTCODE_REQUIRED and self.country_id:
            country_code = self.country.iso_3166_1_a2
            regex = self.POSTCODES_REGEX.get(country_code, None)
            if regex:
                msg = _("Addresses in %(country)s require a valid postcode") % {
                    "country": self.country
                }
                raise exceptions.ValidationError(msg)

        if self.postcode and self.country_id:
            # Ensure postcodes are always uppercase
            postcode = self.postcode.upper().replace(" ", "")
            country_code = self.country.iso_3166_1_a2
            regex = self.POSTCODES_REGEX.get(country_code, None)

            # Validate postcode against regex for the country if available
            if regex and not re.match(regex, postcode):
                msg = _("The postcode '%(postcode)s' is not valid for %(country)s") % {
                    "postcode": self.postcode,
                    "country": self.country,
                }
                raise exceptions.ValidationError({"postcode": [msg]})

    def _update_search_text(self):
        self.search_text = self.join_fields(self.search_fields, separator=" ")

    # Properties

    @property
    def city(self):
        # Common alias
        return self.line4

    @property
    def summary(self):
        """
        Returns a single string summary of the address,
        separating fields using commas.
        """
        return ", ".join(self.active_address_fields())

    @property
    def salutation(self):
        """
        Name (including title)
        """
        return self.join_fields(
            ("title", "first_name", "last_name"), separator=" "
        ).strip()

    @property
    def name(self):
        return self.join_fields(("first_name", "last_name"), separator=" ")

    # Helpers

    def get_field_values(self, fields):
        field_values = []
        for field in fields:
            # Title is special case
            if field == "title":
                value = self.get_title_display()
            elif field == "country":
                try:
                    value = self.country.printable_name
                except exceptions.ObjectDoesNotExist:
                    value = ""
            elif field == "salutation":
                value = self.salutation
            else:
                value = getattr(self, field)
            field_values.append(value)
        return field_values

    def get_address_field_values(self, fields):
        """
        Returns set of field values within the salutation and country.
        """
        field_values = [f.strip() for f in self.get_field_values(fields) if f]
        return field_values

    def generate_hash(self):
        """
        Returns a hash of the address, based on standard set of fields, listed
        out in `hash_fields` property.
        """
        field_values = self.get_address_field_values(self.hash_fields)
        # Python 2 and 3 generates CRC checksum in different ranges, so
        # in order to generate platform-independent value we apply
        # `& 0xffffffff` expression.
        return zlib.crc32(", ".join(field_values).upper().encode("UTF8")) & 0xFFFFFFFF

    def join_fields(self, fields, separator=", "):
        """
        Join a sequence of fields using the specified separator
        """
        field_values = self.get_field_values(fields)
        return separator.join(filter(bool, field_values))

    def populate_alternative_model(self, address_model):
        """
        For populating an address model using the matching fields
        from this one.

        This is used to convert a user address to a shipping address
        as part of the checkout process.
        """
        destination_field_names = [field.name for field in address_model._meta.fields]
        for field_name in [field.name for field in self._meta.fields]:
            if field_name in destination_field_names and field_name != "id":
                setattr(address_model, field_name, getattr(self, field_name))

    def active_address_fields(self):
        """
        Returns the non-empty components of the address, but merging the
        title, first_name and last_name into a single line. It uses fields
        listed out in `base_fields` property.
        """
        return self.get_address_field_values(self.base_fields)


class AbstractCountry(models.Model):
    """
    `ISO 3166 Country Codes <https://www.iso.org/iso-3166-country-codes.html>`_

    The field names are a bit awkward, but kept for backwards compatibility.
    pycountry's syntax of alpha2, alpha3, name and official_name seems sane.
    """

    iso_3166_1_a2 = models.CharField(
        _("ISO 3166-1 alpha-2"), max_length=2, primary_key=True
    )
    iso_3166_1_a3 = models.CharField(_("ISO 3166-1 alpha-3"), max_length=3, blank=True)
    iso_3166_1_numeric = models.CharField(
        _("ISO 3166-1 numeric"), blank=True, max_length=3
    )

    #: The commonly used name; e.g. 'United Kingdom'
    printable_name = models.CharField(_("Country name"), max_length=128, db_index=True)
    #: The full official name of a country
    #: e.g. 'United Kingdom of Great Britain and Northern Ireland'
    name = models.CharField(_("Official name"), max_length=128)

    display_order = models.PositiveSmallIntegerField(
        _("Display order"),
        default=0,
        db_index=True,
        help_text=_("Higher the number, higher the country in the list."),
    )

    is_shipping_country = models.BooleanField(
        _("Is shipping country"), default=False, db_index=True
    )

    class Meta:
        abstract = True
        app_label = "address"
        verbose_name = _("Country")
        verbose_name_plural = _("Countries")
        ordering = (
            "-display_order",
            "printable_name",
        )

    def __str__(self):
        return self.printable_name or self.name

    @property
    def code(self):
        """
        Shorthand for the ISO 3166 Alpha-2 code
        """
        return self.iso_3166_1_a2

    @property
    def numeric_code(self):
        """
        Shorthand for the ISO 3166 numeric code.

        :py:attr:`.iso_3166_1_numeric` used to wrongly be a integer field, but has to
        be padded with leading zeroes. It's since been converted to a char
        field, but the database might still contain non-padded strings. That's
        why the padding is kept.
        """
        return "%.03d" % int(self.iso_3166_1_numeric)


class AbstractShippingAddress(AbstractAddress):
    """
    A shipping address.

    A shipping address should not be edited once the order has been placed -
    it should be read-only after that.

    NOTE:
    ShippingAddress is a model of the order app. But moving it there is tricky
    due to circular import issues that are amplified by get_model/get_class
    calls pre-Django 1.7 to register receivers. So...
    TODO: Once Django 1.6 support is dropped, move AbstractBillingAddress and
    AbstractShippingAddress to the order app, and move
    PartnerAddress to the partner app.
    """

    phone_number = PhoneNumberField(
        _("Phone number"),
        blank=True,
        help_text=_("In case we need to call you about your order"),
    )
    notes = models.TextField(
        blank=True,
        verbose_name=_("Instructions"),
        help_text=_("Tell us anything we should know when delivering your order."),
    )

    class Meta:
        abstract = True
        # ShippingAddress is registered in order/models.py
        app_label = "order"
        verbose_name = _("Shipping address")
        verbose_name_plural = _("Shipping addresses")

    @property
    def order(self):
        """
        Return the order linked to this shipping address
        """
        return self.order_set.first()


class AbstractUserAddress(AbstractShippingAddress):
    """
    A user's address.  A user can have many of these and together they form an
    'address book' of sorts for the user.

    We use a separate model for shipping and billing (even though there will be
    some data duplication) because we don't want shipping/billing addresses
    changed or deleted once an order has been placed.  By having a separate
    model, we allow users the ability to add/edit/delete from their address
    book without affecting orders already placed.
    """

    user = models.ForeignKey(
        AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="addresses",
        verbose_name=_("User"),
    )

    #: Whether this address is the default for shipping
    is_default_for_shipping = models.BooleanField(
        _("Default shipping address?"), default=False
    )

    #: Whether this address should be the default for billing.
    is_default_for_billing = models.BooleanField(
        _("Default billing address?"), default=False
    )

    #: We keep track of the number of times an address has been used
    #: as a shipping address so we can show the most popular ones
    #: first at the checkout.
    num_orders_as_shipping_address = models.PositiveIntegerField(
        _("Number of Orders as Shipping Address"), default=0
    )

    #: Same as previous, but for billing address.
    num_orders_as_billing_address = models.PositiveIntegerField(
        _("Number of Orders as Billing Address"), default=0
    )

    #: A hash is kept to try and avoid duplicate addresses being added
    #: to the address book.
    hash = models.CharField(
        _("Address Hash"), max_length=255, db_index=True, editable=False
    )
    date_created = models.DateTimeField(_("Date Created"), auto_now_add=True)

    def save(self, *args, **kwargs):
        """
        Save a hash of the address fields
        """
        # Save a hash of the address fields so we can check whether two
        # addresses are the same to avoid saving duplicates
        self.hash = self.generate_hash()

        # Ensure that each user only has one default shipping address
        # and billing address
        self._ensure_defaults_integrity()
        super().save(*args, **kwargs)

    def _ensure_defaults_integrity(self):
        if self.is_default_for_shipping:
            self.__class__._default_manager.filter(
                user=self.user, is_default_for_shipping=True
            ).update(is_default_for_shipping=False)
        if self.is_default_for_billing:
            self.__class__._default_manager.filter(
                user=self.user, is_default_for_billing=True
            ).update(is_default_for_billing=False)

    class Meta:
        abstract = True
        app_label = "address"
        verbose_name = _("User address")
        verbose_name_plural = _("User addresses")
        ordering = ["-num_orders_as_shipping_address"]
        unique_together = ("user", "hash")

    def validate_unique(self, exclude=None):
        super().validate_unique(exclude)
        qs = self.__class__.objects.filter(user=self.user, hash=self.generate_hash())
        if self.id:
            qs = qs.exclude(id=self.id)
        if qs.exists():
            raise exceptions.ValidationError(
                {"__all__": [_("This address is already in your address book")]}
            )


class AbstractBillingAddress(AbstractAddress):
    class Meta:
        abstract = True
        # BillingAddress is registered in order/models.py
        app_label = "order"
        verbose_name = _("Billing address")
        verbose_name_plural = _("Billing addresses")

    @property
    def order(self):
        """
        Return the order linked to this shipping address
        """
        return self.order_set.first()


class AbstractPartnerAddress(AbstractAddress):
    """
    A partner can have one or more addresses. This can be useful e.g. when
    determining US tax which depends on the origin of the shipment.
    """

    partner = models.ForeignKey(
        "partner.Partner",
        on_delete=models.CASCADE,
        related_name="addresses",
        verbose_name=_("Partner"),
    )

    class Meta:
        abstract = True
        app_label = "partner"
        verbose_name = _("Partner address")
        verbose_name_plural = _("Partner addresses")



# Generated by Django 4.2.9 on 2024-03-22 19:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("catalogue", "0027_attributeoption_code_attributeoptiongroup_code_and_more"),
        ("offer", "0012_fixedunitdiscountbenefit_alter_benefit_type"),
    ]

    operations = [
        migrations.AddField(
            model_name="range",
            name="excluded_categories",
            field=models.ManyToManyField(
                blank=True,
                help_text="Products with these categories are excluded from the range when Includes all products is checked",
                related_name="excludes",
                to="catalogue.category",
                verbose_name="Excluded Categories",
            ),
        ),
    ]



from django.utils.translation import gettext_lazy as _

from oscar.core.loading import get_class, get_model

ReportGenerator = get_class("dashboard.reports.reports", "ReportGenerator")
ReportCSVFormatter = get_class("dashboard.reports.reports", "ReportCSVFormatter")
ReportHTMLFormatter = get_class("dashboard.reports.reports", "ReportHTMLFormatter")
Basket = get_model("basket", "Basket")


class OpenBasketReportCSVFormatter(ReportCSVFormatter):
    filename_template = "open-baskets-%s-%s.csv"

    def generate_csv(self, response, baskets):
        writer = self.get_csv_writer(response)
        header_row = [
            _("User ID"),
            _("Name"),
            _("Email"),
            _("Basket status"),
            _("Num lines"),
            _("Num items"),
            _("Date of creation"),
            _("Time since creation"),
        ]
        writer.writerow(header_row)

        for basket in baskets:
            if basket.owner:
                row = [
                    basket.owner_id,
                    basket.owner.get_full_name(),
                    basket.owner.email,
                    basket.status,
                    basket.num_lines,
                    basket.num_items,
                    self.format_datetime(basket.date_created),
                    basket.time_since_creation,
                ]
            else:
                row = [
                    basket.owner_id,
                    None,
                    None,
                    basket.status,
                    basket.num_lines,
                    basket.num_items,
                    self.format_datetime(basket.date_created),
                    self.format_timedelta(basket.time_since_creation),
                ]
            writer.writerow(row)

    def filename(self, **kwargs):
        return self.filename_template % (kwargs["start_date"], kwargs["end_date"])


class OpenBasketReportHTMLFormatter(ReportHTMLFormatter):
    filename_template = "oscar/dashboard/reports/partials/open_basket_report.html"


class OpenBasketReportGenerator(ReportGenerator):
    """
    Report of baskets which haven't been submitted yet
    """

    code = "open_baskets"
    description = _("Open baskets")
    date_range_field_name = "date_created"
    queryset = Basket._default_manager.filter(status=Basket.OPEN)

    formatters = {
        "CSV_formatter": OpenBasketReportCSVFormatter,
        "HTML_formatter": OpenBasketReportHTMLFormatter,
    }

    def generate(self):
        additional_data = {"start_date": self.start_date, "end_date": self.end_date}
        return self.formatter.generate_response(self.queryset, **additional_data)


class SubmittedBasketReportCSVFormatter(ReportCSVFormatter):
    filename_template = "submitted_baskets-%s-%s.csv"

    def generate_csv(self, response, baskets):
        writer = self.get_csv_writer(response)
        header_row = [
            _("User ID"),
            _("User"),
            _("Basket status"),
            _("Num lines"),
            _("Num items"),
            _("Date created"),
            _("Time between creation and submission"),
        ]
        writer.writerow(header_row)

        for basket in baskets:
            row = [
                basket.owner_id,
                basket.owner,
                basket.status,
                basket.num_lines,
                basket.num_items,
                self.format_datetime(basket.date_created),
                basket.time_before_submit,
            ]
            writer.writerow(row)

    def filename(self, **kwargs):
        return self.filename_template % (kwargs["start_date"], kwargs["end_date"])


class SubmittedBasketReportHTMLFormatter(ReportHTMLFormatter):
    filename_template = "oscar/dashboard/reports/partials/submitted_basket_report.html"


class SubmittedBasketReportGenerator(ReportGenerator):
    """
    Report of baskets that have been submitted
    """

    code = "submitted_baskets"
    description = _("Submitted baskets")
    date_range_field_name = "date_submitted"
    queryset = Basket._default_manager.filter(status=Basket.SUBMITTED)

    formatters = {
        "CSV_formatter": SubmittedBasketReportCSVFormatter,
        "HTML_formatter": SubmittedBasketReportHTMLFormatter,
    }

    def generate(self):
        additional_data = {"start_date": self.start_date, "end_date": self.end_date}
        return self.formatter.generate_response(self.queryset, **additional_data)



# Generated by Django 4.2.9 on 2024-03-21 16:06

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("order", "0015_orderlinediscount"),
    ]

    operations = [
        migrations.AddField(
            model_name="line",
            name="allocation_cancelled",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="line",
            name="num_allocated",
            field=models.PositiveIntegerField(
                blank=True, null=True, verbose_name="Number allocated"
            ),
        ),
    ]



from django.contrib import admin

from oscar.core.loading import get_model

Order = get_model("order", "Order")
OrderNote = get_model("order", "OrderNote")
OrderStatusChange = get_model("order", "OrderStatusChange")
CommunicationEvent = get_model("order", "CommunicationEvent")
BillingAddress = get_model("order", "BillingAddress")
ShippingAddress = get_model("order", "ShippingAddress")
Line = get_model("order", "Line")
LinePrice = get_model("order", "LinePrice")
ShippingEvent = get_model("order", "ShippingEvent")
ShippingEventType = get_model("order", "ShippingEventType")
PaymentEvent = get_model("order", "PaymentEvent")
PaymentEventType = get_model("order", "PaymentEventType")
PaymentEventQuantity = get_model("order", "PaymentEventQuantity")
LineAttribute = get_model("order", "LineAttribute")
OrderDiscount = get_model("order", "OrderDiscount")
Surcharge = get_model("order", "Surcharge")


class LineInline(admin.TabularInline):
    model = Line
    extra = 0


class OrderAdmin(admin.ModelAdmin):
    raw_id_fields = [
        "user",
        "billing_address",
        "shipping_address",
    ]
    list_display = (
        "number",
        "total_incl_tax",
        "site",
        "user",
        "billing_address",
        "date_placed",
    )
    readonly_fields = (
        "number",
        "basket",
        "total_incl_tax",
        "total_excl_tax",
        "shipping_incl_tax",
        "shipping_excl_tax",
    )
    inlines = [LineInline]


class LineAdmin(admin.ModelAdmin):
    list_display = ("order", "product", "stockrecord", "quantity")


class LinePriceAdmin(admin.ModelAdmin):
    list_display = ("order", "line", "price_incl_tax", "quantity")


class ShippingEventTypeAdmin(admin.ModelAdmin):
    list_display = ("name",)


class PaymentEventQuantityInline(admin.TabularInline):
    model = PaymentEventQuantity
    extra = 0


class PaymentEventAdmin(admin.ModelAdmin):
    list_display = (
        "order",
        "event_type",
        "amount",
        "num_affected_lines",
        "date_created",
    )
    inlines = [PaymentEventQuantityInline]


class PaymentEventTypeAdmin(admin.ModelAdmin):
    pass


class OrderDiscountAdmin(admin.ModelAdmin):
    readonly_fields = (
        "order",
        "category",
        "offer_id",
        "offer_name",
        "voucher_id",
        "voucher_code",
        "amount",
    )
    list_display = ("order", "category", "offer", "voucher", "voucher_code", "amount")


class SurchargeAdmin(admin.ModelAdmin):
    raw_id_fields = ("order",)


admin.site.register(Order, OrderAdmin)
admin.site.register(OrderNote)
admin.site.register(OrderStatusChange)
admin.site.register(ShippingAddress)
admin.site.register(Line, LineAdmin)
admin.site.register(LinePrice, LinePriceAdmin)
admin.site.register(ShippingEvent)
admin.site.register(ShippingEventType, ShippingEventTypeAdmin)
admin.site.register(PaymentEvent, PaymentEventAdmin)
admin.site.register(PaymentEventType, PaymentEventTypeAdmin)
admin.site.register(LineAttribute)
admin.site.register(OrderDiscount, OrderDiscountAdmin)
admin.site.register(CommunicationEvent)
admin.site.register(BillingAddress)
admin.site.register(Surcharge, SurchargeAdmin)



# Generated by Django 2.2.6 on 2020-02-19 09:16

from django.db import migrations, models
import django.db.models.deletion
from django.utils.module_loading import import_string
from django.conf import settings

models_AutoField = import_string(settings.DEFAULT_AUTO_FIELD)


class Migration(migrations.Migration):

    dependencies = [
        ('order', '0008_auto_20190301_1035'),
    ]
    
    replaces=[
        ("order", "0008_surcharge"),
    ]

    operations = [
        migrations.CreateModel(
            name='Surcharge',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128, verbose_name='Name')),
                ('code', models.CharField(max_length=128, verbose_name='Surcharge code')),
                ('incl_tax', models.DecimalField(decimal_places=2, default=0, max_digits=12, verbose_name='Surcharge (inc. tax)')),
                ('excl_tax', models.DecimalField(decimal_places=2, default=0, max_digits=12, verbose_name='Surcharge (excl. tax)')),
                ('order', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='surcharges', to='order.Order', verbose_name='Order')),
            ],
            options={
                'abstract': False,
                'ordering': ['pk'],
                'verbose_name': 'Surcharge',
                'verbose_name_plural': 'Surcharges'
            },
        ),
    ]



# Generated by Django 2.0.13 on 2019-04-30 16:36

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('customer', '0005_auto_20181115_1953'),
    ]

    state_operations = [
        migrations.DeleteModel(
            name='CommunicationEventType',
        ),
        migrations.RemoveField(
            model_name='email',
            name='user',
        ),
        migrations.RemoveField(
            model_name='notification',
            name='recipient',
        ),
        migrations.RemoveField(
            model_name='notification',
            name='sender',
        ),
        migrations.DeleteModel(
            name='Email',
        ),
        migrations.DeleteModel(
            name='Notification',
        ),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=state_operations
        ),
    ]



from django.urls import path
from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarDashboardConfig
from oscar.core.loading import get_class


class ReviewsDashboardConfig(OscarDashboardConfig):
    label = "reviews_dashboard"
    name = "oscar.apps.dashboard.reviews"
    verbose_name = _("Reviews dashboard")

    default_permissions = [
        "is_staff",
    ]

    def configure_permissions(self):
        DashboardPermission = get_class("dashboard.permissions", "DashboardPermission")

        self.permissions_map = {
            "reviews-list": DashboardPermission.get("product_review"),
            "reviews-update": DashboardPermission.get("product_review"),
            "reviews-delete": DashboardPermission.get("product_review"),
        }

    # pylint: disable=attribute-defined-outside-init
    def ready(self):
        self.list_view = get_class("dashboard.reviews.views", "ReviewListView")
        self.update_view = get_class("dashboard.reviews.views", "ReviewUpdateView")
        self.delete_view = get_class("dashboard.reviews.views", "ReviewDeleteView")
        self.configure_permissions()

    def get_urls(self):
        urls = [
            path("", self.list_view.as_view(), name="reviews-list"),
            path("<int:pk>/", self.update_view.as_view(), name="reviews-update"),
            path("<int:pk>/delete/", self.delete_view.as_view(), name="reviews-delete"),
        ]
        return self.post_process_urls(urls)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('voucher', '0001_initial'),
        ('basket', '0002_auto_20140827_1705'),
    ]

    operations = [
        migrations.AddField(
            model_name='basket',
            name='vouchers',
            field=models.ManyToManyField(blank=True, verbose_name='Vouchers', to='voucher.Voucher', null=True),
            preserve_default=True,
        ),
    ]



# Generated by Django 2.2.10 on 2020-07-24 08:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('partner', '0005_auto_20181115_1953'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='stockrecord',
            name='cost_price',
        ),
        migrations.RemoveField(
            model_name='stockrecord',
            name='price_retail',
        ),
        migrations.AlterField(
            model_name='stockrecord',
            name='price_excl_tax',
            field=models.DecimalField(blank=True, decimal_places=2, max_digits=12, null=True, verbose_name='Price'),
        ),
        migrations.RenameField(
            model_name='stockrecord',
            old_name='price_excl_tax',
            new_name='price',
        ),
    ]



# Generated by Django 3.2.9 on 2022-01-25 20:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0024_remove_duplicate_attributes'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='productattribute',
            unique_together={('code', 'product_class')},
        ),
    ]



import factory
from django.contrib.auth import models as auth_models
from django.contrib.sites import models as sites_models

__all__ = ["PermissionFactory", "SiteFactory"]


class PermissionFactory(factory.django.DjangoModelFactory):
    name = "Dummy permission"
    codename = "dummy"

    class Meta:
        model = auth_models.Permission
        django_get_or_create = ("content_type", "codename")


class SiteFactory(factory.django.DjangoModelFactory):
    domain = factory.Sequence(lambda n: "site-%d.oscarcommerce.com" % n)

    class Meta:
        model = sites_models.Site
        django_get_or_create = ("domain",)



from oscar.apps.analytics.abstract_models import (
    AbstractProductRecord,
    AbstractUserProductView,
    AbstractUserRecord,
    AbstractUserSearch,
)
from oscar.core.loading import is_model_registered

__all__ = []


if not is_model_registered("analytics", "ProductRecord"):

    class ProductRecord(AbstractProductRecord):
        pass

    __all__.append("ProductRecord")


if not is_model_registered("analytics", "UserRecord"):

    class UserRecord(AbstractUserRecord):
        pass

    __all__.append("UserRecord")


if not is_model_registered("analytics", "UserProductView"):

    class UserProductView(AbstractUserProductView):
        pass

    __all__.append("UserProductView")


if not is_model_registered("analytics", "UserSearch"):

    class UserSearch(AbstractUserSearch):
        pass

    __all__.append("UserSearch")



from django import forms
from django.conf import settings
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from oscar.core.loading import get_model
from oscar.forms import widgets

ConditionalOffer = get_model("offer", "ConditionalOffer")
Condition = get_model("offer", "Condition")
Benefit = get_model("offer", "Benefit")


def get_offer_type_choices():
    return (("", "---------"),) + tuple(
        choice
        for choice in ConditionalOffer.TYPE_CHOICES
        if choice[0]
        in [
            getattr(ConditionalOffer, const_name)
            for const_name in settings.OSCAR_OFFERS_IMPLEMENTED_TYPES
        ]
    )


class MetaDataForm(forms.ModelForm):
    offer_type = forms.ChoiceField(label=_("Type"), choices=get_offer_type_choices)

    class Meta:
        model = ConditionalOffer
        fields = ("name", "description", "offer_type")

    def clean_offer_type(self):
        data = self.cleaned_data["offer_type"]
        if (
            (self.instance.pk is not None)
            and (self.instance.offer_type == ConditionalOffer.VOUCHER)
            and ("offer_type" in self.changed_data)
            and self.instance.vouchers.exists()
        ):
            raise forms.ValidationError(
                _("This can only be changed if it has no vouchers attached to it")
            )
        return data


class RestrictionsForm(forms.ModelForm):
    start_datetime = forms.DateTimeField(
        widget=widgets.DateTimePickerInput(), label=_("Start date"), required=False
    )
    end_datetime = forms.DateTimeField(
        widget=widgets.DateTimePickerInput(), label=_("End date"), required=False
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["start_datetime"].initial = timezone.now()

    class Meta:
        model = ConditionalOffer
        fields = (
            "start_datetime",
            "end_datetime",
            "max_basket_applications",
            "max_user_applications",
            "max_global_applications",
            "max_discount",
            "priority",
            "exclusive",
            "combinations",
        )

    def clean(self):
        cleaned_data = super().clean()
        start = cleaned_data["start_datetime"]
        end = cleaned_data["end_datetime"]
        if start and end and end < start:
            raise forms.ValidationError(_("The end date must be after the start date"))
        exclusive = cleaned_data["exclusive"]
        combinations = cleaned_data["combinations"]
        if exclusive and combinations:
            raise forms.ValidationError(_("Exclusive offers cannot be combined"))
        return cleaned_data

    def save(self, *args, **kwargs):
        """Store the offer combinations.

        Also, and make sure the combinations are stored on the combine-able
        offers as well.
        """
        instance = super().save(*args, **kwargs)
        if instance.id:
            for offer in instance.combinations.all():
                if offer not in self.cleaned_data["combinations"]:
                    offer.combinations.remove(instance)
            instance.combinations.clear()
            for offer in self.cleaned_data["combinations"]:
                if offer != instance:
                    instance.combinations.add(offer)

            combined_offers = instance.combined_offers
            for offer in combined_offers:
                if offer == instance:
                    continue
                for otheroffer in combined_offers:
                    if offer == otheroffer:
                        continue
                    offer.combinations.add(otheroffer)
        return instance


class ConditionForm(forms.ModelForm):
    custom_condition = forms.ChoiceField(
        required=False, label=_("Custom condition"), choices=()
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        custom_conditions = Condition.objects.all().exclude(proxy_class=None)
        if len(custom_conditions) > 0:
            # Initialise custom_condition field
            choices = [(c.id, str(c)) for c in custom_conditions]
            choices.insert(0, ("", " --------- "))
            self.fields["custom_condition"].choices = choices
            condition = kwargs.get("instance")
            if condition:
                self.fields["custom_condition"].initial = condition.id
        else:
            # No custom conditions and so the type/range/value fields
            # are no longer optional
            for field in ("type", "range", "value"):
                self.fields[field].required = True

    class Meta:
        model = Condition
        fields = ["range", "type", "value"]

    def clean(self):
        data = super().clean()

        # Check that either a condition has been entered or a custom condition
        # has been chosen
        if not any(data.values()):
            raise forms.ValidationError(
                _(
                    "Please either choose a range, type and value OR "
                    "select a custom condition"
                )
            )

        if data["custom_condition"]:
            if data.get("range") or data.get("type") or data.get("value"):
                raise forms.ValidationError(
                    _(
                        "No other options can be set if you are using a "
                        "custom condition"
                    )
                )
        elif not data.get("type"):
            raise forms.ValidationError(
                _(
                    "Please either choose a range, type and value OR "
                    "select a custom condition"
                )
            )

        return data

    def save(self, *args, **kwargs):
        # We don't save a new model if a custom condition has been chosen,
        # we simply return the instance that has been chosen
        if self.cleaned_data["custom_condition"]:
            return Condition.objects.get(id=self.cleaned_data["custom_condition"])
        return super().save(*args, **kwargs)


class BenefitForm(forms.ModelForm):
    custom_benefit = forms.ChoiceField(
        required=False, label=_("Custom incentive"), choices=()
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        custom_benefits = Benefit.objects.all().exclude(proxy_class=None)
        if len(custom_benefits) > 0:
            # Initialise custom_benefit field
            choices = [(c.id, str(c)) for c in custom_benefits]
            choices.insert(0, ("", " --------- "))
            self.fields["custom_benefit"].choices = choices
            benefit = kwargs.get("instance")
            if benefit:
                self.fields["custom_benefit"].initial = benefit.id
        else:
            # No custom benefit and so the type fields
            # are no longer optional
            self.fields["type"].required = True

    class Meta:
        model = Benefit
        fields = ["range", "type", "value", "max_affected_items"]

    def clean(self):
        data = super().clean()

        # Check that either a benefit has been entered or a custom benfit
        # has been chosen
        if not any(data.values()):
            raise forms.ValidationError(
                _(
                    "Please either choose a range, type and value OR "
                    "select a custom incentive"
                )
            )

        if data["custom_benefit"]:
            if data.get("range") or data.get("type") or data.get("value"):
                raise forms.ValidationError(
                    _(
                        "No other options can be set if you are using a "
                        "custom incentive"
                    )
                )
        elif not data.get("type"):
            raise forms.ValidationError(
                _(
                    "Please either choose a range, type and value OR "
                    "select a custom incentive"
                )
            )

        return data

    def save(self, *args, **kwargs):
        # We don't save a new model if a custom benefit has been chosen,
        # we simply return the instance that has been chosen
        if self.cleaned_data["custom_benefit"]:
            return Benefit.objects.get(id=self.cleaned_data["custom_benefit"])
        return super().save(*args, **kwargs)


class OfferSearchForm(forms.Form):
    name = forms.CharField(required=False, label=_("Offer name"))
    is_active = forms.NullBooleanField(
        required=False, label=_("Is active?"), widget=widgets.NullBooleanSelect
    )
    offer_type = forms.ChoiceField(
        required=False, label=_("Offer type"), choices=get_offer_type_choices
    )
    has_vouchers = forms.NullBooleanField(
        required=False, label=_("Has vouchers?"), widget=widgets.NullBooleanSelect
    )
    voucher_code = forms.CharField(required=False, label=_("Voucher code"))

    basic_fields = [
        "name",
        "is_active",
    ]

    @property
    def is_voucher_offer_type(self):
        return self.is_bound and (
            self.cleaned_data["offer_type"] == ConditionalOffer.VOUCHER
        )



# pylint: disable=attribute-defined-outside-init
import csv

from django.conf import settings
from django.contrib import messages
from django.db import transaction
from django.db.models import Count
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse, reverse_lazy
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django.views import generic

from oscar.core.loading import get_class, get_model
from oscar.core.utils import slugify
from oscar.views import sort_queryset

VoucherForm = get_class("dashboard.vouchers.forms", "VoucherForm")
VoucherSetForm = get_class("dashboard.vouchers.forms", "VoucherSetForm")
VoucherSetSearchForm = get_class("dashboard.vouchers.forms", "VoucherSetSearchForm")
VoucherSearchForm = get_class("dashboard.vouchers.forms", "VoucherSearchForm")
Voucher = get_model("voucher", "Voucher")
VoucherSet = get_model("voucher", "VoucherSet")
OrderDiscount = get_model("order", "OrderDiscount")


class VoucherListView(generic.ListView):
    model = Voucher
    context_object_name = "vouchers"
    template_name = "oscar/dashboard/vouchers/voucher_list.html"
    form_class = VoucherSearchForm
    paginate_by = settings.OSCAR_DASHBOARD_ITEMS_PER_PAGE

    def get_queryset(self):
        self.search_filters = []
        qs = self.model._default_manager.all()
        qs = qs.annotate(num_offers=Count("offers", distinct=True))
        qs = sort_queryset(
            qs,
            self.request,
            ["num_basket_additions", "num_orders", "num_offers", "date_created"],
            "-date_created",
        )

        # If form is not submitted, perform a default filter, and return early
        if not self.request.GET:
            self.form = self.form_class(initial={"in_set": False})
            # This form is exactly the same as the other one, apart from having
            # fields with different IDs, so that they are unique within the page
            # (non-unique field IDs also break Select2)
            self.advanced_form = self.form_class(
                initial={"in_set": False}, auto_id="id_advanced_%s"
            )
            self.search_filters.append(_("Not in a set"))
            return qs.filter(voucher_set__isnull=True)

        self.form = self.form_class(self.request.GET)
        # This form is exactly the same as the other one, apart from having
        # fields with different IDs, so that they are unique within the page
        # (non-unique field IDs also break Select2)
        self.advanced_form = self.form_class(self.request.GET, auto_id="id_advanced_%s")
        if not all([self.form.is_valid(), self.advanced_form.is_valid()]):
            return qs

        name = self.form.cleaned_data["name"]
        code = self.form.cleaned_data["code"]
        offer_name = self.form.cleaned_data["offer_name"]
        is_active = self.form.cleaned_data["is_active"]
        in_set = self.form.cleaned_data["in_set"]
        has_offers = self.form.cleaned_data["has_offers"]

        if name:
            qs = qs.filter(name__icontains=name)
            self.search_filters.append(_('Name matches "%s"') % name)
        if code:
            qs = qs.filter(code=code)
            self.search_filters.append(_('Code is "%s"') % code)
        if offer_name:
            qs = qs.filter(offers__name__icontains=offer_name)
            self.search_filters.append(_('Offer name matches "%s"') % offer_name)
        if is_active is not None:
            now = timezone.now()
            if is_active:
                qs = qs.filter(start_datetime__lte=now, end_datetime__gte=now)
                self.search_filters.append(_("Is active"))
            else:
                qs = qs.filter(end_datetime__lt=now)
                self.search_filters.append(_("Is inactive"))
        if in_set is not None:
            qs = qs.filter(voucher_set__isnull=not in_set)
            self.search_filters.append(_("In a set") if in_set else _("Not in a set"))
        if has_offers is not None:
            qs = qs.filter(offers__isnull=not has_offers).distinct()
            self.search_filters.append(
                _("Has offers") if has_offers else _("Has no offers")
            )

        return qs

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["form"] = self.form
        ctx["advanced_form"] = self.advanced_form
        ctx["search_filters"] = self.search_filters
        return ctx


class VoucherCreateView(generic.CreateView):
    model = Voucher
    template_name = "oscar/dashboard/vouchers/voucher_form.html"
    form_class = VoucherForm
    success_url = reverse_lazy("dashboard:voucher-list")

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["title"] = _("Create voucher")
        return ctx

    def get_initial(self):
        initial = super().get_initial()
        initial["start_datetime"] = timezone.now()
        return initial

    @transaction.atomic()
    def form_valid(self, form):
        response = super().form_valid(form)
        self.object.offers.add(*form.cleaned_data["offers"])
        return response

    def get_success_url(self):
        messages.success(self.request, _("Voucher created"))
        return super().get_success_url()


class VoucherStatsView(generic.DetailView):
    model = Voucher
    template_name = "oscar/dashboard/vouchers/voucher_detail.html"
    context_object_name = "voucher"

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        discounts = OrderDiscount.objects.filter(voucher_id=self.object.id)
        discounts = discounts.order_by("-order__date_placed")
        ctx["discounts"] = discounts
        return ctx


class VoucherUpdateView(generic.UpdateView):
    template_name = "oscar/dashboard/vouchers/voucher_form.html"
    context_object_name = "voucher"
    model = Voucher
    form_class = VoucherForm
    success_url = reverse_lazy("dashboard:voucher-list")

    def dispatch(self, request, *args, **kwargs):
        voucher_set = self.get_object().voucher_set
        if voucher_set is not None:
            messages.warning(
                request, _("The voucher can only be edited as part of its set")
            )
            return redirect("dashboard:voucher-set-update", pk=voucher_set.pk)
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["title"] = self.object.name
        return ctx

    def get_initial(self):
        initial = super().get_initial()
        initial["offers"] = self.object.offers.all()
        return initial

    @transaction.atomic()
    def form_valid(self, form):
        response = super().form_valid(form)
        self.object.offers.set(form.cleaned_data["offers"])
        return response

    def get_success_url(self):
        messages.success(self.request, _("Voucher updated"))
        return super().get_success_url()


class VoucherDeleteView(generic.DeleteView):
    model = Voucher
    template_name = "oscar/dashboard/vouchers/voucher_delete.html"
    context_object_name = "voucher"

    @transaction.atomic
    def delete(self, request, *args, **kwargs):
        response = super().delete(request, *args, **kwargs)
        if self.object.voucher_set is not None:
            self.object.voucher_set.update_count()
        return response

    def post(self, request, *args, **kwargs):
        return self.delete(request, *args, **kwargs)

    def get_success_url(self):
        messages.warning(self.request, _("Voucher deleted"))
        if self.object.voucher_set is not None:
            return reverse(
                "dashboard:voucher-set-detail",
                kwargs={"pk": self.object.voucher_set.pk},
            )
        else:
            return reverse("dashboard:voucher-list")


class VoucherSetCreateView(generic.CreateView):
    model = VoucherSet
    template_name = "oscar/dashboard/vouchers/voucher_set_form.html"
    form_class = VoucherSetForm

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["title"] = _("Create voucher set")
        return ctx

    def get_initial(self):
        initial = super().get_initial()
        initial["start_datetime"] = timezone.now()
        return initial

    def get_success_url(self):
        messages.success(self.request, _("Voucher set created"))
        return reverse("dashboard:voucher-set-list")


class VoucherSetUpdateView(generic.UpdateView):
    template_name = "oscar/dashboard/vouchers/voucher_set_form.html"
    model = VoucherSet
    context_object_name = "voucher_set"
    form_class = VoucherSetForm

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["title"] = self.object.name
        return ctx

    def get_initial(self):
        initial = super().get_initial()
        # All vouchers in the set have the same "usage" and "offers", so we use
        # the first one
        voucher = self.object.vouchers.first()
        if voucher is not None:
            initial["usage"] = voucher.usage
            initial["offers"] = voucher.offers.all()
        return initial

    def get_success_url(self):
        messages.success(self.request, _("Voucher updated"))
        return reverse("dashboard:voucher-set-detail", kwargs={"pk": self.object.pk})


class VoucherSetDetailView(generic.ListView):
    model = Voucher
    context_object_name = "vouchers"
    template_name = "oscar/dashboard/vouchers/voucher_set_detail.html"
    form_class = VoucherSetSearchForm
    paginate_by = 50

    def dispatch(self, request, *args, **kwargs):
        self.voucher_set = get_object_or_404(VoucherSet, pk=kwargs["pk"])
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        self.search_filters = []
        qs = self.model.objects.filter(voucher_set=self.voucher_set).order_by(
            "-date_created"
        )

        qs = sort_queryset(
            qs,
            self.request,
            ["num_basket_additions", "num_orders", "date_created"],
            "-date_created",
        )

        # If form not submitted, return early
        is_form_submitted = "name" in self.request.GET or "code" in self.request.GET
        if not is_form_submitted:
            self.form = self.form_class()
            return qs

        self.form = self.form_class(self.request.GET)
        if not self.form.is_valid():
            return qs

        data = self.form.cleaned_data
        if data["code"]:
            qs = qs.filter(code__icontains=data["code"])
            self.search_filters.append(_('Code matches "%s"') % data["code"])

        return qs

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["voucher_set"] = self.voucher_set
        ctx["form"] = self.form
        ctx["search_filters"] = self.search_filters
        return ctx


class VoucherSetListView(generic.ListView):
    model = VoucherSet
    context_object_name = "voucher_sets"
    template_name = "oscar/dashboard/vouchers/voucher_set_list.html"
    paginate_by = settings.OSCAR_DASHBOARD_ITEMS_PER_PAGE

    def get_queryset(self):
        qs = self.model.objects.all().order_by("-date_created")
        qs = sort_queryset(
            qs,
            self.request,
            ["num_basket_additions", "num_orders", "date_created"],
            "-date_created",
        )
        return qs

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        description = _("Voucher sets")
        ctx["description"] = description
        return ctx


class VoucherSetDownloadView(generic.DetailView):
    template_name = "oscar/dashboard/vouchers/voucher_set_form.html"
    model = VoucherSet
    form_class = VoucherSetForm

    def get(self, request, *args, **kwargs):
        voucher_set = self.get_object()

        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = 'attachment; filename="%s.csv"' % slugify(
            voucher_set.name
        )

        writer = csv.writer(response)
        for code in voucher_set.vouchers.values_list("code", flat=True):
            writer.writerow([code])

        return response


class VoucherSetDeleteView(generic.DeleteView):
    model = VoucherSet
    template_name = "oscar/dashboard/vouchers/voucher_set_delete.html"
    context_object_name = "voucher_set"

    def get_success_url(self):
        messages.warning(self.request, _("Voucher set deleted"))
        return reverse("dashboard:voucher-set-list")



from collections import defaultdict, namedtuple

from django.contrib import messages
from django.template.loader import render_to_string

from oscar.core.loading import get_class, get_model
from oscar.core.decorators import deprecated

Applicator = get_class("offer.applicator", "Applicator")
ConditionalOffer = get_model("offer", "ConditionalOffer")

DiscountApplication = namedtuple(
    "DiscountApplication", ["amount", "quantity", "incl_tax", "offer"]
)


class BasketMessageGenerator(object):
    new_total_template_name = "oscar/basket/messages/new_total.html"
    offer_lost_template_name = "oscar/basket/messages/offer_lost.html"
    offer_gained_template_name = "oscar/basket/messages/offer_gained.html"

    def get_new_total_messages(self, basket, include_buttons=True):
        new_total_messages = []
        # We use the 'include_buttons' parameter to determine whether to show the
        # 'Checkout now' buttons.  We don't want to show these on the basket page.
        msg = render_to_string(
            self.new_total_template_name,
            {"basket": basket, "include_buttons": include_buttons},
        )
        new_total_messages.append((messages.INFO, msg))

        return new_total_messages

    def get_offer_lost_messages(self, offers_before, offers_after):
        offer_messages = []
        for offer_id in set(offers_before).difference(offers_after):
            offer = offers_before[offer_id]
            msg = render_to_string(self.offer_lost_template_name, {"offer": offer})
            offer_messages.append((messages.WARNING, msg))
        return offer_messages

    def get_offer_gained_messages(self, offers_before, offers_after):
        offer_messages = []
        for offer_id in set(offers_after).difference(offers_before):
            offer = offers_after[offer_id]
            msg = render_to_string(self.offer_gained_template_name, {"offer": offer})
            offer_messages.append((messages.SUCCESS, msg))
        return offer_messages

    def get_offer_messages(self, offers_before, offers_after):
        offer_messages = []
        offer_messages.extend(self.get_offer_lost_messages(offers_before, offers_after))
        offer_messages.extend(
            self.get_offer_gained_messages(offers_before, offers_after)
        )
        return offer_messages

    def get_messages(self, basket, offers_before, offers_after, include_buttons=True):
        message_list = []
        message_list.extend(self.get_offer_messages(offers_before, offers_after))
        message_list.extend(self.get_new_total_messages(basket, include_buttons))
        return message_list

    def apply_messages(self, request, offers_before):
        """
        Set flash messages triggered by changes to the basket
        """
        # Re-apply offers to see if any new ones are now available
        request.basket.reset_offer_applications()
        Applicator().apply(request.basket, request.user, request)
        offers_after = request.basket.applied_offers()

        for level, msg in self.get_messages(
            request.basket, offers_before, offers_after
        ):
            messages.add_message(request, level, msg, extra_tags="safe noicon")


class LineOfferConsumer(object):
    """
    facade for marking basket lines as consumed by
    any or a specific offering.

    historically oscar marks a line as consumed if any
    offer is applied to it, but more complicated scenarios
    are possible if we mark the line as being consumed by
    specific offers.

    this allows combining i.e. multiple vouchers, vouchers
    with special session discounts, etc.
    """

    def __init__(self, line):
        self._line = line
        self._offers = dict()
        self._affected_quantity = 0
        self._consumptions = defaultdict(int)

    def _cache(self, offer):
        self._offers[offer.pk] = offer

    def _update_affected_quantity(self, quantity):
        available = int(self._line.quantity - self._affected_quantity)
        num_consumed = min(available, quantity)
        self._affected_quantity += num_consumed
        return num_consumed

    # public
    def consume(self, quantity: int, offer=None):
        """
        mark a basket line as consumed by an offer

        :param int quantity: the number of items on the line affected
        :param offer: the offer to mark the line
        :type offer: ConditionalOffer or None
        :return: the number of items actually consumed
        :rtype: int

        if offer is None, the specified quantity of items on this
        basket line is consumed for *any* offer, else only for the
        specified offer.
        """
        if offer:
            self._cache(offer)
            available = self.available(offer)

        num_consumed = self._update_affected_quantity(quantity)
        if offer:
            num_consumed = min(available, quantity)
            self._consumptions[offer.pk] += num_consumed
        return num_consumed

    @deprecated
    def consumed(self, offer=None):
        return self.num_consumed(offer)

    def num_consumed(self, offer=None):
        """
        check how many items on this line have been
        consumed by an offer

        :param offer: the offer to check
        :type offer: ConditionalOffer or None
        :return: the number of items marked as consumed
        :rtype: int

        if offer is not None, only the number of items marked
        with the specified ConditionalOffer are returned

        """
        if not offer:
            return self._affected_quantity
        return int(self._consumptions[offer.pk])

    @property
    def consumers(self):
        return [x for x in self._offers.values() if self.num_consumed(x)]

    def available(self, offer=None) -> int:
        """
        check how many items are available for offer

        :param offer: the offer to check
        :type offer: ConditionalOffer or None
        :return: the number of items available for offer
        :rtype: int
        """
        max_affected_items = self._line.quantity

        if offer and isinstance(offer, ConditionalOffer):
            applied = [x for x in self.consumers if x != offer]

            if offer.exclusive:
                for a in applied:
                    if a.exclusive:
                        if any(
                            [
                                a.priority > offer.priority,
                                a.priority == offer.priority and a.id != offer.id,
                            ]
                        ):
                            # Exclusive offers cannot be applied if any other exclusive
                            # offer with higher priority is active already.
                            max_affected_items = max_affected_items - self.num_consumed(
                                a
                            )
                            if max_affected_items == 0:
                                return 0

                    else:
                        # Exclusive offers cannot be applied if any other offers are
                        # active already.
                        return 0

            # find any *other* exclusive offers
            elif any([x.exclusive for x in applied]):
                return 0

            # check for applied offers allowing restricted combinations
            for x in applied:
                check = offer.combinations.count() or x.combinations.count()
                if check and offer not in x.combined_offers:
                    return 0

        return max_affected_items - self.num_consumed(offer)


class LineDiscountRegistry(LineOfferConsumer):
    def __init__(self, line):
        super().__init__(line)
        self._discounts = []
        self._discount_excl_tax = None
        self._discount_incl_tax = None

    def discount(self, amount, quantity, incl_tax=True, offer=None):
        self._discounts.append(DiscountApplication(amount, quantity, incl_tax, offer))
        self.consume(quantity, offer=offer)
        if incl_tax:
            self._discount_incl_tax = None
        else:
            self._discount_excl_tax = None

    @property
    def excl_tax(self):
        if self._discount_excl_tax is None:
            self._discount_excl_tax = sum(
                [d.amount for d in self._discounts if not d.incl_tax], 0
            )
        return self._discount_excl_tax

    @property
    def incl_tax(self):
        if self._discount_incl_tax is None:
            self._discount_incl_tax = sum(
                [d.amount for d in self._discounts if d.incl_tax], 0
            )
        return self._discount_incl_tax

    @property
    def total(self):
        return sum([d.amount for d in self._discounts], 0)

    def all(self):
        return self._discounts

    def __iter__(self):
        return iter(self._discounts)



# pylint: disable=W0401, W0614
from oscar.apps.offer.abstract_models import (
    AbstractBenefit,
    AbstractCondition,
    AbstractConditionalOffer,
    AbstractRange,
    AbstractRangeProduct,
    AbstractRangeProductFileUpload,
)
from oscar.apps.offer.results import (
    SHIPPING_DISCOUNT,
    ZERO_DISCOUNT,
    BasketDiscount,
    PostOrderAction,
    ShippingDiscount,
)
from oscar.core.loading import is_model_registered

__all__ = [
    "BasketDiscount",
    "ShippingDiscount",
    "PostOrderAction",
    "SHIPPING_DISCOUNT",
    "ZERO_DISCOUNT",
]


if not is_model_registered("offer", "ConditionalOffer"):

    class ConditionalOffer(AbstractConditionalOffer):
        pass

    __all__.append("ConditionalOffer")


if not is_model_registered("offer", "Benefit"):

    class Benefit(AbstractBenefit):
        pass

    __all__.append("Benefit")


if not is_model_registered("offer", "Condition"):

    class Condition(AbstractCondition):
        pass

    __all__.append("Condition")


if not is_model_registered("offer", "Range"):

    class Range(AbstractRange):
        pass

    __all__.append("Range")


if not is_model_registered("offer", "RangeProduct"):

    class RangeProduct(AbstractRangeProduct):
        pass

    __all__.append("RangeProduct")


if not is_model_registered("offer", "RangeProductFileUpload"):

    class RangeProductFileUpload(AbstractRangeProductFileUpload):
        pass

    __all__.append("RangeProductFileUpload")

# Import the benefits and the conditions. Required after initializing the
# parent models to allow overriding them

from oscar.apps.offer.benefits import *
from oscar.apps.offer.conditions import *

from oscar.apps.offer.benefits import __all__ as benefit_classes
from oscar.apps.offer.conditions import __all__ as condition_classes

__all__.extend(benefit_classes)
__all__.extend(condition_classes)



from django.contrib import admin

from oscar.core.loading import get_model

Source = get_model("payment", "Source")
Transaction = get_model("payment", "Transaction")
SourceType = get_model("payment", "SourceType")
Bankcard = get_model("payment", "Bankcard")


class SourceAdmin(admin.ModelAdmin):
    list_display = (
        "order",
        "source_type",
        "amount_allocated",
        "amount_debited",
        "balance",
        "reference",
    )


class BankcardAdmin(admin.ModelAdmin):
    list_display = ("number", "card_type", "expiry_month")


admin.site.register(Source, SourceAdmin)
admin.site.register(SourceType)
admin.site.register(Transaction)
admin.site.register(Bankcard, BankcardAdmin)



from django.conf import settings


def is_solr_supported():
    try:
        return "Solr" in settings.HAYSTACK_CONNECTIONS["default"]["ENGINE"]
    except (KeyError, AttributeError):
        return False


def is_elasticsearch_supported():
    try:
        return "Elasticsearch" in settings.HAYSTACK_CONNECTIONS["default"]["ENGINE"]
    except (KeyError, AttributeError):
        return False



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-04-20 05:51
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0009_slugfield_noop'),
    ]

    operations = [
        migrations.AddField(
            model_name='productattributevalue',
            name='value_multi_option',
            field=models.ManyToManyField(blank=True, related_name='multi_valued_attribute_values', to='catalogue.AttributeOption', verbose_name='Value multi option'),
        ),
        migrations.AlterField(
            model_name='productattribute',
            name='option_group',
            field=models.ForeignKey(blank=True, help_text='Select an option group if using type "Option" or "Multi Option"', null=True, on_delete=django.db.models.deletion.CASCADE, to='catalogue.AttributeOptionGroup', verbose_name='Option Group'),
        ),
        migrations.AlterField(
            model_name='productattribute',
            name='type',
            field=models.CharField(choices=[('text', 'Text'), ('integer', 'Integer'), ('boolean', 'True / False'), ('float', 'Float'), ('richtext', 'Rich Text'), ('date', 'Date'), ('option', 'Option'), ('multi_option', 'Multi Option'), ('entity', 'Entity'), ('file', 'File'), ('image', 'Image')], default='text', max_length=20, verbose_name='Type'),
        ),
    ]



from django.urls import path, re_path
from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarDashboardConfig
from oscar.core.loading import get_class


class CommunicationsDashboardConfig(OscarDashboardConfig):
    label = "communications_dashboard"
    name = "oscar.apps.dashboard.communications"
    verbose_name = _("Communications dashboard")

    default_permissions = [
        "is_staff",
    ]

    def configure_permissions(self):
        DashboardPermission = get_class("dashboard.permissions", "DashboardPermission")

        self.permissions_map = {
            "comms-list": DashboardPermission.get("communication_event_type"),
            "comms-update": DashboardPermission.get("communication_event_type"),
        }

    # pylint: disable=attribute-defined-outside-init
    def ready(self):
        self.list_view = get_class("dashboard.communications.views", "ListView")
        self.update_view = get_class("dashboard.communications.views", "UpdateView")
        self.configure_permissions()

    def get_urls(self):
        urls = [
            path("", self.list_view.as_view(), name="comms-list"),
            re_path(
                r"^(?P<slug>[\w-]+)/$", self.update_view.as_view(), name="comms-update"
            ),
        ]
        return self.post_process_urls(urls)



from django.contrib import admin

from oscar.core.loading import get_model

Voucher = get_model("voucher", "Voucher")
VoucherApplication = get_model("voucher", "VoucherApplication")


class VoucherAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "code",
        "usage",
        "num_basket_additions",
        "num_orders",
        "total_discount",
    )
    readonly_fields = ("num_basket_additions", "num_orders", "total_discount")
    fieldsets = (
        (None, {"fields": ("name", "code", "usage", "start_datetime", "end_datetime")}),
        ("Benefit", {"fields": ("offers",)}),
        ("Usage", {"fields": ("num_basket_additions", "num_orders", "total_discount")}),
    )


class VoucherApplicationAdmin(admin.ModelAdmin):
    list_display = ("voucher", "user", "order", "date_created")
    readonly_fields = ("voucher", "user", "order")


admin.site.register(Voucher, VoucherAdmin)
admin.site.register(VoucherApplication, VoucherApplicationAdmin)



import logging
from datetime import timedelta

from django.core.management.base import BaseCommand
from django.utils.timezone import now

from oscar.core.loading import get_model

ProductAlert = get_model("customer", "ProductAlert")

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Command to remove all stale unconfirmed alerts
    """

    help = "Check unconfirmed alerts and clean them up"

    def add_arguments(self, parser):
        parser.add_argument(
            "--days",
            dest="days",
            default=0,
            help="cleanup alerts older then DAYS from now.",
        )
        parser.add_argument(
            "--hours",
            dest="hours",
            default=0,
            help="cleanup alerts older then HOURS from now.",
        )

    def handle(self, *args, **options):
        """
        Generate a threshold date from the input options or 24 hours
        if no options specified. All alerts that have the
        status ``UNCONFIRMED`` and have been created before the
        threshold date will be removed assuming that the emails
        are wrong or the customer changed their mind.
        """
        delta = timedelta(days=int(options["days"]), hours=int(options["hours"]))
        if not delta:
            delta = timedelta(hours=24)

        threshold_date = now() - delta

        logger.info(
            "Deleting unconfirmed alerts older than %s",
            threshold_date.strftime("%Y-%m-%d %H:%M"),
        )

        qs = ProductAlert.objects.filter(
            status=ProductAlert.UNCONFIRMED, date_created__lt=threshold_date
        )
        logger.info("Found %d stale alerts to delete", qs.count())
        qs.delete()



import logging

from django.contrib.sites.models import Site
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponseRedirect
from django.urls import NoReverseMatch, reverse

from oscar.apps.checkout.signals import post_checkout
from oscar.core.loading import get_class, get_model

OrderCreator = get_class("order.utils", "OrderCreator")
OrderDispatcher = get_class("order.utils", "OrderDispatcher")
CheckoutSessionMixin = get_class("checkout.session", "CheckoutSessionMixin")
BillingAddress = get_model("order", "BillingAddress")
ShippingAddress = get_model("order", "ShippingAddress")
OrderNumberGenerator = get_class("order.utils", "OrderNumberGenerator")
PaymentEventType = get_model("order", "PaymentEventType")
PaymentEvent = get_model("order", "PaymentEvent")
PaymentEventQuantity = get_model("order", "PaymentEventQuantity")
UserAddress = get_model("address", "UserAddress")
Basket = get_model("basket", "Basket")

# Standard logger for checkout events
logger = logging.getLogger("oscar.checkout")


class OrderPlacementMixin(CheckoutSessionMixin):
    """
    Mixin which provides functionality for placing orders.

    Any view class which needs to place an order should use this mixin.
    """

    # Any payment sources should be added to this list as part of the
    # handle_payment method.  If the order is placed successfully, then
    # they will be persisted. We need to have the order instance before the
    # payment sources can be saved.
    _payment_sources = None

    # Any payment events should be added to this list as part of the
    # handle_payment method.
    _payment_events = None

    view_signal = post_checkout

    # Payment handling methods
    # ------------------------

    def handle_payment(self, order_number, total, **kwargs):
        """
        Handle any payment processing and record payment sources and events.

        This method is designed to be overridden within your project.  The
        default is to do nothing as payment is domain-specific.

        This method is responsible for handling payment and recording the
        payment sources (using the add_payment_source method) and payment
        events (using add_payment_event) so they can be
        linked to the order when it is saved later on.
        """

    def add_payment_source(self, source):
        """
        Record a payment source for this order
        """
        if self._payment_sources is None:
            self._payment_sources = []
        self._payment_sources.append(source)

    def add_payment_event(self, event_type_name, amount, reference=""):
        """
        Record a payment event for creation once the order is placed
        """
        event_type, __ = PaymentEventType.objects.get_or_create(name=event_type_name)
        # We keep a local cache of (unsaved) payment events
        if self._payment_events is None:
            self._payment_events = []
        event = PaymentEvent(event_type=event_type, amount=amount, reference=reference)
        self._payment_events.append(event)

    # Placing order methods
    # ---------------------

    def generate_order_number(self, basket):
        """
        Return a new order number
        """
        return OrderNumberGenerator().order_number(basket)

    def handle_order_placement(
        self,
        order_number,
        user,
        basket,
        shipping_address,
        shipping_method,
        shipping_charge,
        billing_address,
        order_total,
        surcharges=None,
        **kwargs,
    ):
        """
        Write out the order models and return the appropriate HTTP response

        We deliberately pass the basket in here as the one tied to the request
        isn't necessarily the correct one to use in placing the order.  This
        can happen when a basket gets frozen.
        """
        order = self.place_order(
            order_number=order_number,
            user=user,
            basket=basket,
            shipping_address=shipping_address,
            shipping_method=shipping_method,
            shipping_charge=shipping_charge,
            order_total=order_total,
            billing_address=billing_address,
            surcharges=surcharges,
            **kwargs,
        )
        basket.submit()
        return self.handle_successful_order(order)

    def place_order(
        self,
        order_number,
        user,
        basket,
        shipping_address,
        shipping_method,
        shipping_charge,
        order_total,
        billing_address=None,
        surcharges=None,
        **kwargs,
    ):
        """
        Writes the order out to the DB including the payment models
        """
        # Create saved shipping address instance from passed in unsaved
        # instance
        shipping_address = self.create_shipping_address(user, shipping_address)

        # We pass the kwargs as they often include the billing address form
        # which will be needed to save a billing address.
        billing_address = self.create_billing_address(
            user, billing_address, shipping_address, **kwargs
        )

        if "status" not in kwargs:
            # pylint: disable=assignment-from-none
            status = self.get_initial_order_status(basket)
        else:
            status = kwargs.pop("status")

        if "request" not in kwargs:
            request = getattr(self, "request", None)
        else:
            request = kwargs.pop("request")

        order = OrderCreator().place_order(
            user=user,
            order_number=order_number,
            basket=basket,
            shipping_address=shipping_address,
            shipping_method=shipping_method,
            shipping_charge=shipping_charge,
            total=order_total,
            billing_address=billing_address,
            status=status,
            request=request,
            surcharges=surcharges,
            **kwargs,
        )
        self.save_payment_details(order)
        return order

    def create_shipping_address(self, user, shipping_address):
        """
        Create and return the shipping address for the current order.

        Compared to self.get_shipping_address(), ShippingAddress is saved and
        makes sure that appropriate UserAddress exists.
        """
        # For an order that only contains items that don't require shipping we
        # won't have a shipping address, so we have to check for it.
        if not shipping_address:
            return None
        shipping_address.save()
        if user.is_authenticated:
            self.update_address_book(user, shipping_address)
        return shipping_address

    def update_address_book(self, user, addr):
        """
        Update the user's address book based on the new shipping address
        """
        try:
            user_addr = user.addresses.get(hash=addr.generate_hash())
        except ObjectDoesNotExist:
            # Create a new user address
            user_addr = UserAddress(user=user)
            addr.populate_alternative_model(user_addr)
        if isinstance(addr, ShippingAddress):
            user_addr.num_orders_as_shipping_address += 1
        if isinstance(addr, BillingAddress):
            user_addr.num_orders_as_billing_address += 1
        user_addr.save()

    # pylint: disable=unused-argument
    def create_billing_address(
        self, user, billing_address=None, shipping_address=None, **kwargs
    ):
        """
        Saves any relevant billing data (e.g. a billing address).
        """
        if not billing_address:
            return None
        billing_address.save()
        if user.is_authenticated:
            self.update_address_book(user, billing_address)
        return billing_address

    def save_payment_details(self, order):
        """
        Saves all payment-related details. This could include a billing
        address, payment sources and any order payment events.
        """
        self.save_payment_events(order)
        self.save_payment_sources(order)

    def save_payment_events(self, order):
        """
        Saves any relevant payment events for this order
        """
        if not self._payment_events:
            return
        for event in self._payment_events:
            event.order = order
            event.save()
            for line in order.lines.all():
                PaymentEventQuantity.objects.create(
                    event=event, line=line, quantity=line.quantity
                )

    def save_payment_sources(self, order):
        """
        Saves any payment sources used in this order.

        When the payment sources are created, the order model does not exist
        and so they need to have it set before saving.
        """
        if not self._payment_sources:
            return
        for source in self._payment_sources:
            source.order = order
            source.save()

    # pylint: disable=unused-argument
    def get_initial_order_status(self, basket):
        return None

    # Post-order methods
    # ------------------

    def handle_successful_order(self, order):
        """
        Handle the various steps required after an order has been successfully
        placed.

        Override this view if you want to perform custom actions when an
        order is submitted.
        """
        # Send confirmation message (normally an email)
        self.send_order_placed_email(order)

        # Flush all session data
        self.checkout_session.flush()

        # Save order id in session so thank-you page can load it
        self.request.session["checkout_order_id"] = order.id

        response = HttpResponseRedirect(self.get_success_url())
        self.send_signal(self.request, response, order)
        return response

    def send_signal(self, request, response, order):
        self.view_signal.send(
            sender=self,
            order=order,
            user=request.user,
            request=request,
            response=response,
        )

    def get_success_url(self):
        return reverse("checkout:thank-you")

    def send_order_placed_email(self, order):
        extra_context = self.get_message_context(order)
        dispatcher = OrderDispatcher(logger=logger)
        dispatcher.send_order_placed_email_for_user(order, extra_context)

    def get_message_context(self, order):
        ctx = {
            "user": self.request.user,
            "order": order,
            "lines": order.lines.all(),
            "request": self.request,
        }

        # Attempt to add the order status URL to the email template ctx.
        try:
            if self.request.user.is_authenticated:
                path = reverse("customer:order", kwargs={"order_number": order.number})
            else:
                path = reverse(
                    "customer:anon-order",
                    kwargs={
                        "order_number": order.number,
                        "hash": order.verification_hash(),
                    },
                )
        except NoReverseMatch:
            # We don't care that much if we can't resolve the URL
            pass
        else:
            ctx["status_path"] = path

            # status_url is deprecated, see https://github.com/django-oscar/django-oscar/issues/3826
            site = Site.objects.get_current(self.request)
            ctx["status_url"] = "http://%s%s" % (site.domain, path)
        return ctx

    # Basket helpers
    # --------------

    def get_submitted_basket(self):
        basket_id = self.checkout_session.get_submitted_basket_id()
        return Basket._default_manager.get(pk=basket_id)

    def freeze_basket(self, basket):
        """
        Freeze the basket so it can no longer be modified
        """
        # We freeze the basket to prevent it being modified once the payment
        # process has started.  If your payment fails, then the basket will
        # need to be "unfrozen".  We also store the basket ID in the session
        # so the it can be retrieved by multistage checkout processes.
        basket.freeze()

    def restore_frozen_basket(self):
        """
        Restores a frozen basket as the sole OPEN basket.  Note that this also
        merges in any new products that have been added to a basket that has
        been created while payment.
        """
        try:
            fzn_basket = self.get_submitted_basket()
        except Basket.DoesNotExist:
            # Strange place.  The previous basket stored in the session does
            # not exist.
            pass
        else:
            fzn_basket.thaw()
            if self.request.basket.id != fzn_basket.id:
                fzn_basket.merge(self.request.basket)
                # Use same strategy as current request basket
                fzn_basket.strategy = self.request.basket.strategy
                self.request.basket = fzn_basket



import logging

from django.core.management.base import BaseCommand

from oscar.core.loading import get_class

Importer = get_class("catalogue.utils", "Importer")

logger = logging.getLogger("oscar.catalogue.import")


class Command(BaseCommand):
    help = "For importing product images from a folder"

    def add_arguments(self, parser):
        parser.add_argument("path", help="/path/to/folder")

        parser.add_argument(
            "--filename",
            dest="filename",
            default="upc",
            help="Product field to lookup from image filename",
        )

    def handle(self, *args, **options):
        logger.info("Starting image import")
        dirname = options["path"]
        importer = Importer(logger, field=options.get("filename"))
        importer.handle(dirname)



from django import forms
from django.utils.translation import gettext_lazy as _

from oscar.core.loading import get_model

Vote = get_model("reviews", "vote")
ProductReview = get_model("reviews", "productreview")


class ProductReviewForm(forms.ModelForm):
    name = forms.CharField(label=_("Name"), required=True)
    email = forms.EmailField(label=_("Email"), required=True)

    def __init__(self, product, *args, user=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.instance.product = product
        if user and user.is_authenticated:
            self.instance.user = user
            del self.fields["name"]
            del self.fields["email"]

    class Meta:
        model = ProductReview
        fields = ("title", "score", "body", "name", "email")


class VoteForm(forms.ModelForm):
    class Meta:
        model = Vote
        fields = ("delta",)

    def __init__(self, review, user, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instance.review = review
        self.instance.user = user

    @property
    def is_up_vote(self):
        return self.cleaned_data["delta"] == Vote.UP

    @property
    def is_down_vote(self):
        return self.cleaned_data["delta"] == Vote.DOWN


class SortReviewsForm(forms.Form):
    SORT_BY_SCORE = "score"
    SORT_BY_RECENCY = "recency"
    SORT_REVIEWS_BY_CHOICES = (
        (SORT_BY_SCORE, _("Score")),
        (SORT_BY_RECENCY, _("Recency")),
    )

    sort_by = forms.ChoiceField(
        choices=SORT_REVIEWS_BY_CHOICES,
        label=_("Sort by"),
        initial=SORT_BY_SCORE,
        required=False,
    )



from django.dispatch import receiver

from oscar.apps.catalogue.signals import product_viewed
from oscar.core.loading import get_class

CustomerHistoryManager = get_class("customer.history", "CustomerHistoryManager")


# pylint: disable=unused-argument
@receiver(product_viewed)
def receive_product_view(sender, product, user, request, response, **kwargs):
    """
    Receiver to handle viewing single product pages

    Requires the request and response objects due to dependence on cookies
    """
    return CustomerHistoryManager.update(product, request, response)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('shipping', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='orderanditemcharges',
            name='countries',
            field=models.ManyToManyField(to='address.Country', verbose_name='Countries', blank=True),
        ),
        migrations.AlterField(
            model_name='weightbased',
            name='countries',
            field=models.ManyToManyField(to='address.Country', verbose_name='Countries', blank=True),
        ),
    ]



from oscar.core.loading import get_model

Notification = get_model("communication", "Notification")


def notifications(request):
    ctx = {}
    if getattr(request, "user", None) and request.user.is_authenticated:
        num_unread = Notification.objects.filter(
            recipient=request.user, date_read=None
        ).count()
        ctx["num_unread_notifications"] = num_unread
    return ctx



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-06-09 18:02
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0011_auto_20170422_1355'),
    ]

    operations = [
        migrations.AddField(
            model_name='productattributevalue',
            name='value_datetime',
            field=models.DateTimeField(blank=True, null=True, verbose_name='DateTime'),
        ),
        migrations.AlterField(
            model_name='productattribute',
            name='type',
            field=models.CharField(choices=[('text', 'Text'), ('integer', 'Integer'), ('boolean', 'True / False'), ('float', 'Float'), ('richtext', 'Rich Text'), ('date', 'Date'), ('datetime', 'Datetime'), ('option', 'Option'), ('multi_option', 'Multi Option'), ('entity', 'Entity'), ('file', 'File'), ('image', 'Image')], default='text', max_length=20, verbose_name='Type'),
        ),
    ]



from django import forms

from oscar.core.loading import get_model


class WeightBasedForm(forms.ModelForm):
    class Meta:
        model = get_model("shipping", "WeightBased")
        fields = ["name", "description", "default_weight", "countries"]


class WeightBandForm(forms.ModelForm):
    def __init__(self, method, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instance.method = method

    class Meta:
        model = get_model("shipping", "WeightBand")
        fields = ("upper_limit", "charge")



# Generated by Django 4.2.11 on 2024-06-14 15:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("order", "0017_set_num_allocated_to_quantity"),
    ]

    operations = [
        migrations.AlterField(
            model_name="line",
            name="num_allocated",
            field=models.PositiveIntegerField(verbose_name="Number allocated"),
        ),
    ]



# pylint: disable=W0622
from django.core.checks import Error, Warning, register
from django.conf import settings
from django.db import connection


def turned_on_materialised_views():
    return (
        hasattr(settings, "OSCAR_CATALOGUE_USE_POSTGRES_MATERIALISED_VIEWS")
        and settings.OSCAR_CATALOGUE_USE_POSTGRES_MATERIALISED_VIEWS
    )


def is_postgres():
    return connection.vendor == "postgresql"


@register()
def startup_check(*args, **kwargs):
    errors = []

    if (
        hasattr(settings, "OSCAR_PRODUCT_SEARCH_HANDLER")
        and settings.OSCAR_PRODUCT_SEARCH_HANDLER is not None
    ):
        errors.append(
            Error(
                "The OSCAR_PRODUCT_SEARCH_HANDLER is removed since django-oscar==3.2.4.",
                hint="Use the new class based haystack views instead located in search.views. Any customizations that has been done to the search handler should be moved there.",
                id="django-oscar.E001",
            )
        )

    if turned_on_materialised_views() and not is_postgres():
        errors.append(
            Warning(
                "OSCAR_CATALOGUE_USE_POSTGRES_MATERIALISED_VIEWS is enabled but PostgreSQL is not detected. Materialized views are not functioning and fallback queries are being used instead.",
                hint="Either switch your database backend to PostgreSQL or disable the OSCAR_CATALOGUE_USE_POSTGRES_MATERIALISED_VIEWS setting in your configuration.",
                id="django-oscar.W001",
            )
        )
    return errors


def use_productcategory_materialised_view():
    return is_postgres() and turned_on_materialised_views()



# pylint: disable=wildcard-import, unused-wildcard-import
from oscar.core.loading import is_model_registered

from .abstract_models import *

__all__ = []


if not is_model_registered("communication", "Email"):

    class Email(AbstractEmail):
        pass

    __all__.append("Email")


if not is_model_registered("communication", "CommunicationEventType"):

    class CommunicationEventType(AbstractCommunicationEventType):
        pass

    __all__.append("CommunicationEventType")


if not is_model_registered("communication", "Notification"):

    class Notification(AbstractNotification):
        pass

    __all__.append("Notification")



# Generated by Django 2.0.13 on 2019-04-30 16:35

from django.db import migrations


class Migration(migrations.Migration):

    atomic = False

    dependencies = [
        ('communication', '0001_initial'),
        ('order', '0008_auto_20190301_1035'),
    ]

    operations = [
        # Alter the names of the tables migrated from the customer app
        # to the Django defaults.
        migrations.AlterModelTable(
            name='communicationeventtype',
            table=None,
        ),
        migrations.AlterModelTable(
            name='email',
            table=None,
        ),
        migrations.AlterModelTable(
            name='notification',
            table=None,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.utils.module_loading import import_string
from django.conf import settings
from decimal import Decimal

models_AutoField = import_string(settings.DEFAULT_AUTO_FIELD)



class Migration(migrations.Migration):

    dependencies = [
        ('order', '0001_initial'),
        ('offer', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Voucher',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(verbose_name='Name', max_length=128, help_text='This will be shown in the checkout and basket once the voucher is entered')),
                ('code', models.CharField(max_length=128, verbose_name='Code', unique=True, db_index=True, help_text='Case insensitive / No spaces allowed')),
                ('usage', models.CharField(default='Multi-use', max_length=128, verbose_name='Usage', choices=[('Single use', 'Can be used once by one customer'), ('Multi-use', 'Can be used multiple times by multiple customers'), ('Once per customer', 'Can only be used once per customer')])),
                ('start_datetime', models.DateTimeField(verbose_name='Start datetime')),
                ('end_datetime', models.DateTimeField(verbose_name='End datetime')),
                ('num_basket_additions', models.PositiveIntegerField(default=0, verbose_name='Times added to basket')),
                ('num_orders', models.PositiveIntegerField(default=0, verbose_name='Times on orders')),
                ('total_discount', models.DecimalField(default=Decimal('0.00'), max_digits=12, decimal_places=2, verbose_name='Total discount')),
                ('date_created', models.DateField(auto_now_add=True)),
                ('offers', models.ManyToManyField(related_name='vouchers', verbose_name='Offers', to='offer.ConditionalOffer')),
            ],
            options={
                'verbose_name_plural': 'Vouchers',
                'get_latest_by': 'date_created',
                'verbose_name': 'Voucher',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='VoucherApplication',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateField(auto_now_add=True, verbose_name='Date Created')),
                ('order', models.ForeignKey(verbose_name='Order', to='order.Order', on_delete=models.CASCADE)),
                ('user', models.ForeignKey(null=True, verbose_name='User', to=settings.AUTH_USER_MODEL, blank=True, on_delete=models.CASCADE)),
                ('voucher', models.ForeignKey(verbose_name='Voucher', related_name='applications', to='voucher.Voucher', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name_plural': 'Voucher Applications',
                'verbose_name': 'Voucher Application',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
    ]



from django.db.models import F
from django.utils.translation import gettext_lazy as _
from django_tables2 import A, Column, LinkColumn, TemplateColumn

from oscar.core.loading import get_class

DashboardTable = get_class("dashboard.tables", "DashboardTable")


class UserTable(DashboardTable):
    check = TemplateColumn(
        template_name="oscar/dashboard/users/user_row_checkbox.html",
        verbose_name=" ",
        orderable=False,
    )
    email = LinkColumn("dashboard:user-detail", args=[A("id")], accessor="email")
    name = Column(accessor="get_full_name", order_by=("last_name", "first_name"))
    active = Column(accessor="is_active")
    staff = Column(accessor="is_staff")
    date_registered = Column(accessor="date_joined")
    num_orders = Column(
        accessor="userrecord__num_orders", default=0, verbose_name=_("Number of Orders")
    )
    actions = TemplateColumn(
        template_name="oscar/dashboard/users/user_row_actions.html", verbose_name=" "
    )

    icon = "fas fa-users"

    class Meta(DashboardTable.Meta):
        template_name = "oscar/dashboard/users/table.html"

    def order_num_orders(self, queryset, is_descending):
        """
        User record is created only once a user places an order, all such records must
        be at the end of listing, when sorting in descending order, and at the start
        when sorting in ascending.
        """
        if is_descending:
            order_by = F("userrecord__num_orders").desc(nulls_last=True)
        else:
            order_by = F("userrecord__num_orders").asc(nulls_first=True)

        return (queryset.order_by(order_by), True)



# Generated by Django 3.0.10 on 2020-10-05 07:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0020_auto_20200801_0817'),
    ]

    operations = [
        migrations.AddField(
            model_name='category',
            name='meta_description',
            field=models.TextField(blank=True, null=True, verbose_name='Meta description'),
        ),
        migrations.AddField(
            model_name='category',
            name='meta_title',
            field=models.CharField(blank=True, null=True, max_length=255, verbose_name='Meta title'),
        ),
        migrations.AddField(
            model_name='product',
            name='meta_description',
            field=models.TextField(blank=True, null=True, verbose_name='Meta description'),
        ),
        migrations.AddField(
            model_name='product',
            name='meta_title',
            field=models.CharField(blank=True, null=True, max_length=255, verbose_name='Meta title'),
        ),
    ]



from django.conf import settings
from django.db.models import SlugField as DjangoSlugField


class SlugField(DjangoSlugField):
    def __init__(self, *args, **kwargs):
        # not override parameter if it was passed explicitly,
        # so passed parameters takes precedence over the setting
        if settings.OSCAR_SLUG_ALLOW_UNICODE:
            kwargs.setdefault("allow_unicode", settings.OSCAR_SLUG_ALLOW_UNICODE)

        super().__init__(*args, **kwargs)



# Generated by Django 5.2.4 on 2025-07-15 12:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("catalogue", "0031_productcategoryhierarchy"),
    ]

    operations = [
        migrations.AddField(
            model_name="category",
            name="exclude_from_menu",
            field=models.BooleanField(
                db_index=True,
                default=False,
                help_text="Exclude this category from the menu.",
                verbose_name="Exclude from menu",
            ),
        ),
        migrations.AddField(
            model_name="category",
            name="long_description",
            field=models.TextField(blank=True, verbose_name="Long description"),
        ),
    ]



import re

from django import forms
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.forms.models import ModelChoiceIterator
from django.forms.widgets import FileInput
from django.utils import formats
from django.utils.encoding import force_str
from django.utils.translation import gettext as _


class ImageInput(FileInput):
    """
    Widget providing a input element for file uploads based on the
    Django ``FileInput`` element. It hides the actual browser-specific
    input element and shows the available image for images that have
    been previously uploaded. Selecting the image will open the file
    dialog and allow for selecting a new or replacing image file.
    """

    template_name = "oscar/forms/widgets/image_input_widget.html"

    def __init__(self, attrs=None):
        if not attrs:
            attrs = {}
        attrs["accept"] = "image/*"
        super().__init__(attrs=attrs)

    def get_context(self, name, value, attrs):
        ctx = super().get_context(name, value, attrs)

        ctx["image_url"] = ""
        if value and not isinstance(value, InMemoryUploadedFile):
            # can't display images that aren't stored - pass empty string to context
            ctx["image_url"] = value

        ctx["image_id"] = "%s-image" % ctx["widget"]["attrs"]["id"]
        return ctx


class WYSIWYGTextArea(forms.Textarea):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("attrs", {})
        kwargs["attrs"].setdefault("class", "")
        kwargs["attrs"]["class"] += " wysiwyg"
        super().__init__(*args, **kwargs)


# pylint: disable=W0622
def datetime_format_to_js_date_format(datetime_format):
    """
    Convert a Python datetime format to a date format suitable for use with
    the JS date picker we use.
    """
    datetime_format = datetime_format.split()[0]
    return datetime_format_to_js_datetime_format(datetime_format)


# pylint: disable=W0622
def datetime_format_to_js_time_format(datetime_format):
    """
    Convert a Python datetime format to a time format suitable for use with the
    JS time picker we use.
    """
    try:
        datetime_format = datetime_format.split()[1]
    except IndexError:
        pass
    converted = datetime_format
    replacements = {
        "%H": "HH",
        "%I": "hh",
        "%M": "mm",
        "%S": "ss",
    }
    for search, replace in replacements.items():
        converted = converted.replace(search, replace)
    return converted.strip()


def datetime_format_to_js_datetime_format(datetime_format):
    """
    Convert a Python datetime format to a time format suitable for use with
    the datetime picker we use, http://momentjs.com/docs/#/displaying/format/.
    """
    converted = datetime_format
    replacements = {
        "%Y": "YYYY",
        "%y": "YY",
        "%m": "MM",
        "%d": "DD",
        "%H": "HH",
        "%I": "hh",
        "%M": "mm",
        "%S": "ss",
    }
    for search, replace in replacements.items():
        converted = converted.replace(search, replace)

    return converted.strip()


def datetime_format_to_js_input_mask(datetime_format):
    # taken from
    # http://stackoverflow.com/questions/15175142/how-can-i-do-multiple-substitutions-using-regex-in-python
    def multiple_replace(replacements, datetime_format):
        # Create a regular expression  from the dictionary keys
        regex = re.compile("(%s)" % "|".join(map(re.escape, replacements.keys())))

        # For each match, look-up corresponding value in dictionary
        return regex.sub(
            lambda mo: replacements[mo.string[mo.start() : mo.end()]], datetime_format
        )

    replacements = {
        "%Y": "yyyy",
        "%y": "yy",
        "%m": "mm",
        "%d": "dd",
        "%H": "HH",
        "%I": "hh",
        "%M": "MM",
        "%S": "ss",
    }
    return multiple_replace(replacements, datetime_format).strip()


class DateTimeWidgetMixin(object):
    template_name = "oscar/forms/widgets/date_time_picker.html"

    def get_format(self):
        return self.format or formats.get_format(self.format_key)[0]

    def build_attrs(self, base_attrs, extra_attrs=None):
        attrs = super().build_attrs(base_attrs, extra_attrs)
        attrs["data-inputmask"] = "'alias': 'datetime', 'inputFormat': '{mask}'".format(
            mask=datetime_format_to_js_input_mask(self.get_format())
        )
        return attrs


class TimePickerInput(DateTimeWidgetMixin, forms.TimeInput):
    """
    A widget that passes the date format to the JS date picker in a data
    attribute.
    """

    format_key = "TIME_INPUT_FORMATS"

    def get_context(self, name, value, attrs):
        ctx = super().get_context(name, value, attrs)
        ctx["div_attrs"] = {
            "data-oscarWidget": "time",
            "data-timeFormat": datetime_format_to_js_time_format(self.get_format()),
        }
        ctx["icon_classes"] = "far fa-clock"
        return ctx


class DatePickerInput(DateTimeWidgetMixin, forms.DateInput):
    """
    A widget that passes the date format to the JS date picker in a data
    attribute.
    """

    format_key = "DATE_INPUT_FORMATS"

    def get_context(self, name, value, attrs):
        ctx = super().get_context(name, value, attrs)
        ctx["div_attrs"] = {
            "data-oscarWidget": "date",
            "data-dateFormat": datetime_format_to_js_date_format(self.get_format()),
        }
        ctx["icon_classes"] = "far fa-calendar-alt"
        return ctx


class DateTimePickerInput(DateTimeWidgetMixin, forms.DateTimeInput):
    """
    A widget that passes the datetime format to the JS datetime picker in a
    data attribute.

    It also removes seconds by default. However this only works with widgets
    without localize=True.

    For localized widgets refer to
    https://docs.djangoproject.com/en/1.11/topics/i18n/formatting/#creating-custom-format-files
    instead to override the format.
    """

    format_key = "DATETIME_INPUT_FORMATS"

    def __init__(self, *args, **kwargs):
        include_seconds = kwargs.pop("include_seconds", False)
        super().__init__(*args, **kwargs)

        if not include_seconds and self.format:
            self.format = re.sub(":?%S", "", self.format)

    def get_context(self, name, value, attrs):
        ctx = super().get_context(name, value, attrs)
        ctx["div_attrs"] = {
            "data-oscarWidget": "datetime",
            "data-datetimeFormat": datetime_format_to_js_datetime_format(
                self.get_format()
            ),
        }
        ctx["icon_classes"] = "far fa-calendar-alt"
        return ctx


class AdvancedSelect(forms.Select):
    """
    Customised Select widget that allows a list of disabled values to be passed
    to the constructor.  Django's default Select widget doesn't allow this so
    we have to override the render_option method and add a section that checks
    for whether the widget is disabled.
    """

    def __init__(self, attrs=None, choices=(), disabled_values=()):
        self.disabled_values = set(force_str(v) for v in disabled_values)
        super().__init__(attrs, choices)

    def create_option(
        self, name, value, label, selected, index, subindex=None, attrs=None
    ):
        option = super().create_option(
            name, value, label, selected, index, subindex, attrs
        )
        if force_str(value) in self.disabled_values:
            option["attrs"]["disabled"] = True
        return option


class RemoteSelect(forms.Select):
    """
    Somewhat reusable widget that allows AJAX lookups in combination with
    select2.
    Requires setting the URL of a lookup view either as class attribute or when
    constructing
    """

    lookup_url = None

    def __init__(self, *args, **kwargs):
        self.lookup_url = kwargs.pop("lookup_url", self.lookup_url)
        if self.lookup_url is None:
            raise ValueError("RemoteSelect requires a lookup URL")

        super().__init__(*args, **kwargs)

    def build_attrs(self, *args, **kwargs):
        attrs = super().build_attrs(*args, **kwargs)
        attrs.update(
            {
                "data-ajax-url": self.lookup_url,
                "data-multiple": "multiple" if self.allow_multiple_selected else "",
                "data-required": "required" if self.is_required else "",
            }
        )
        return attrs

    def optgroups(self, name, value, attrs=None):
        """
        Borrowed liberally from https://github.com/applegrew/django-select2

        For model choice fields, return only the selected options. The rest
        will be populated with AJAX.
        """

        default = (None, [], 0)
        groups = [default]
        has_selected = False
        selected_choices = {force_str(v) for v in value}
        if not self.is_required and not self.allow_multiple_selected:
            default[1].append(self.create_option(name, "", "", False, 0))

        # If thi is not a model choice field then just return all choices
        if not isinstance(self.choices, ModelChoiceIterator):
            return super().optgroups(name, value, attrs=attrs)

        selected_choices = {
            c for c in selected_choices if c not in self.choices.field.empty_values
        }
        choices = (
            (obj.pk, force_str(obj))
            for obj in self.choices.queryset.filter(pk__in=selected_choices)
        )
        for option_value, option_label in choices:
            selected = str(option_value) in value and (
                has_selected is False or self.allow_multiple_selected
            )
            if selected is True and has_selected is False:
                has_selected = True
            index = len(default[1])
            subgroup = default[1]
            subgroup.append(
                self.create_option(
                    name, option_value, option_label, selected_choices, index
                )
            )

        return groups


class MultipleRemoteSelect(RemoteSelect):
    allow_multiple_selected = True


class NullBooleanSelect(forms.NullBooleanSelect):
    """
    Customised NullBooleanSelect widget that gives the "unknown" choice a more
    meaningful label than the default of "Unknown".
    """

    def __init__(self, attrs=None):
        super().__init__(attrs)
        self.choices = (
            ("unknown", _("---------")),
            ("true", _("Yes")),
            ("false", _("No")),
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-04-22 12:55
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0010_auto_20170420_0439'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='productimage',
            unique_together=set([]),
        ),
    ]



import logging
from urllib.parse import quote

from django import http
from django.conf import settings
from django.contrib import messages
from django.contrib.auth import login
from django.shortcuts import redirect
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext as _
from django.views import generic

from oscar.core.loading import get_class, get_classes, get_model

from . import signals

ShippingAddressForm, ShippingMethodForm, GatewayForm = get_classes(
    "checkout.forms", ["ShippingAddressForm", "ShippingMethodForm", "GatewayForm"]
)
UserAddressForm = get_class("address.forms", "UserAddressForm")
Repository = get_class("shipping.repository", "Repository")
RedirectRequired, UnableToTakePayment, PaymentError = get_classes(
    "payment.exceptions", ["RedirectRequired", "UnableToTakePayment", "PaymentError"]
)
UnableToPlaceOrder = get_class("order.exceptions", "UnableToPlaceOrder")
OrderPlacementMixin = get_class("checkout.mixins", "OrderPlacementMixin")
CheckoutSessionMixin = get_class("checkout.session", "CheckoutSessionMixin")
NoShippingRequired = get_class("shipping.methods", "NoShippingRequired")
Order = get_model("order", "Order")
ShippingAddress = get_model("order", "ShippingAddress")
UserAddress = get_model("address", "UserAddress")
Country = get_model("address", "Country")

# Standard logger for checkout events
logger = logging.getLogger("oscar.checkout")


class IndexView(CheckoutSessionMixin, generic.FormView):
    """
    First page of the checkout.  We prompt user to either sign in, or
    to proceed as a guest (where we still collect their email address).
    """

    template_name = "oscar/checkout/gateway.html"
    form_class = GatewayForm
    success_url = reverse_lazy("checkout:shipping-address")
    pre_conditions = ["check_basket_is_not_empty", "check_basket_is_valid"]

    def get(self, request, *args, **kwargs):
        # We redirect immediately to shipping address stage if the user is
        # signed in.
        if request.user.is_authenticated:
            # We raise a signal to indicate that the user has entered the
            # checkout process so analytics tools can track this event.
            signals.start_checkout.send_robust(sender=self, request=request)
            return self.get_success_response()
        return super().get(request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        email = self.checkout_session.get_guest_email()
        if email:
            kwargs["initial"] = {
                "username": email,
            }
        return kwargs

    def form_valid(self, form):
        if form.is_guest_checkout() or form.is_new_account_checkout():
            email = form.cleaned_data["username"]
            self.checkout_session.set_guest_email(email)

            # We raise a signal to indicate that the user has entered the
            # checkout process by specifying an email address.
            signals.start_checkout.send_robust(
                sender=self, request=self.request, email=email
            )

            if form.is_new_account_checkout():
                messages.info(
                    self.request,
                    _(
                        "Create your account and then you will be redirected "
                        "back to the checkout process"
                    ),
                )
                self.success_url = "%s?next=%s&email=%s" % (
                    reverse("customer:register"),
                    reverse("checkout:shipping-address"),
                    quote(email),
                )
        else:
            user = form.get_user()
            login(self.request, user)

            # We raise a signal to indicate that the user has entered the
            # checkout process.
            signals.start_checkout.send_robust(sender=self, request=self.request)

        return redirect(self.get_success_url())

    def get_success_response(self):
        return redirect(self.get_success_url())


# ================
# SHIPPING ADDRESS
# ================


class ShippingAddressView(CheckoutSessionMixin, generic.FormView):
    """
    Determine the shipping address for the order.

    The default behaviour is to display a list of addresses from the users's
    address book, from which the user can choose one to be their shipping
    address.  They can add/edit/delete these USER addresses.  This address will
    be automatically converted into a SHIPPING address when the user checks
    out.

    Alternatively, the user can enter a SHIPPING address directly which will be
    saved in the session and later saved as ShippingAddress model when the
    order is successfully submitted.
    """

    template_name = "oscar/checkout/shipping_address.html"
    form_class = ShippingAddressForm
    success_url = reverse_lazy("checkout:shipping-method")
    pre_conditions = [
        "check_basket_is_not_empty",
        "check_basket_is_valid",
        "check_user_email_is_captured",
    ]
    skip_conditions = ["skip_unless_basket_requires_shipping"]

    def get_initial(self):
        initial = self.checkout_session.new_shipping_address_fields()
        if initial:
            initial = initial.copy()
            # Convert the primary key stored in the session into a Country
            # instance
            try:
                initial["country"] = Country.objects.get(
                    iso_3166_1_a2=initial.pop("country_id")
                )
            except Country.DoesNotExist:
                # Hmm, the previously selected Country no longer exists. We
                # ignore this.
                pass
        return initial

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        if self.request.user.is_authenticated:
            # Look up address book data
            ctx["addresses"] = self.get_available_addresses()
        return ctx

    def get_available_addresses(self):
        # Include only addresses where the country is flagged as valid for
        # shipping. Also, use ordering to ensure the default address comes
        # first.
        return self.request.user.addresses.filter(
            country__is_shipping_country=True
        ).order_by("-is_default_for_shipping")

    def post(self, request, *args, **kwargs):
        # Check if a shipping address was selected directly (e.g. no form was
        # filled in)
        if self.request.user.is_authenticated and "address_id" in self.request.POST:
            address = UserAddress._default_manager.get(
                pk=self.request.POST["address_id"], user=self.request.user
            )
            action = self.request.POST.get("action", None)
            if action == "ship_to":
                # User has selected a previous address to ship to
                self.checkout_session.ship_to_user_address(address)
                return redirect(self.get_success_url())
            else:
                return http.HttpResponseBadRequest()
        else:
            return super().post(request, *args, **kwargs)

    def form_valid(self, form):
        # Store the address details in the session and redirect to next step
        address_fields = dict(
            (k, v) for (k, v) in form.instance.__dict__.items() if not k.startswith("_")
        )
        self.checkout_session.ship_to_new_address(address_fields)
        return super().form_valid(form)


class UserAddressUpdateView(CheckoutSessionMixin, generic.UpdateView):
    """
    Update a user address
    """

    template_name = "oscar/checkout/user_address_form.html"
    form_class = UserAddressForm
    success_url = reverse_lazy("checkout:shipping-address")

    def get_queryset(self):
        return self.request.user.addresses.all()

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.request.user
        return kwargs

    def get_success_url(self):
        messages.info(self.request, _("Address saved"))
        return super().get_success_url()


class UserAddressDeleteView(CheckoutSessionMixin, generic.DeleteView):
    """
    Delete an address from a user's address book.
    """

    template_name = "oscar/checkout/user_address_delete.html"
    success_url = reverse_lazy("checkout:shipping-address")

    def get_queryset(self):
        return self.request.user.addresses.all()

    def get_success_url(self):
        messages.info(self.request, _("Address deleted"))
        return super().get_success_url()


# ===============
# Shipping method
# ===============


# pylint: disable=attribute-defined-outside-init
class ShippingMethodView(CheckoutSessionMixin, generic.FormView):
    """
    View for allowing a user to choose a shipping method.

    Shipping methods are largely domain-specific and so this view
    will commonly need to be subclassed and customised.

    The default behaviour is to load all the available shipping methods
    using the shipping Repository.  If there is only 1, then it is
    automatically selected.  Otherwise, a page is rendered where
    the user can choose the appropriate one.
    """

    template_name = "oscar/checkout/shipping_methods.html"
    form_class = ShippingMethodForm
    pre_conditions = [
        "check_basket_is_not_empty",
        "check_basket_is_valid",
        "check_user_email_is_captured",
    ]
    success_url = reverse_lazy("checkout:payment-method")

    def post(self, request, *args, **kwargs):
        self._methods = self.get_available_shipping_methods()
        return super().post(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        # These skip and pre conditions can't easily be factored out into the
        # normal pre-conditions as they do more than run a test and then raise
        # an exception on failure.

        # Check that shipping is required at all
        if not request.basket.is_shipping_required():
            # No shipping required - we store a special code to indicate so.
            self.checkout_session.use_shipping_method(NoShippingRequired().code)
            return self.get_success_response()

        # Check that shipping address has been completed
        if not self.checkout_session.is_shipping_address_set():
            messages.error(request, _("Please choose a shipping address"))
            return redirect("checkout:shipping-address")

        # Save shipping methods as instance var as we need them both here
        # and when setting the context vars.
        self._methods = self.get_available_shipping_methods()
        if len(self._methods) == 0:
            # No shipping methods available for given address
            messages.warning(
                request,
                _(
                    "Shipping is unavailable for your chosen address - please "
                    "choose another"
                ),
            )
            return redirect("checkout:shipping-address")
        elif len(self._methods) == 1:
            # Only one shipping method - set this and redirect onto the next
            # step
            self.checkout_session.use_shipping_method(self._methods[0].code)
            return self.get_success_response()

        # Must be more than one available shipping method, we present them to
        # the user to make a choice.
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs["methods"] = self._methods
        return kwargs

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["methods"] = self._methods
        return kwargs

    def get_available_shipping_methods(self):
        """
        Returns all applicable shipping method objects for a given basket.
        """
        # Shipping methods can depend on the user, the contents of the basket
        # and the shipping address (so we pass all these things to the
        # repository).  I haven't come across a scenario that doesn't fit this
        # system.
        return Repository().get_shipping_methods(
            basket=self.request.basket,
            user=self.request.user,
            shipping_addr=self.get_shipping_address(self.request.basket),
            request=self.request,
        )

    def form_valid(self, form):
        # Save the code for the chosen shipping method in the session
        # and continue to the next step.
        self.checkout_session.use_shipping_method(form.cleaned_data["method_code"])
        return self.get_success_response()

    def form_invalid(self, form):
        messages.error(
            self.request, _("Your submitted shipping method is not permitted")
        )
        return super().form_invalid(form)

    def get_success_response(self):
        return redirect(self.get_success_url())


# ==============
# Payment method
# ==============


class PaymentMethodView(CheckoutSessionMixin, generic.TemplateView):
    """
    View for a user to choose which payment method(s) they want to use.

    This would include setting allocations if payment is to be split
    between multiple sources. It's not the place for entering sensitive details
    like bankcard numbers though - that belongs on the payment details view.
    """

    pre_conditions = [
        "check_basket_is_not_empty",
        "check_basket_is_valid",
        "check_user_email_is_captured",
        "check_shipping_data_is_captured",
    ]
    skip_conditions = ["skip_unless_payment_is_required"]
    success_url = reverse_lazy("checkout:payment-details")

    def get(self, request, *args, **kwargs):
        # By default we redirect straight onto the payment details view. Shops
        # that require a choice of payment method may want to override this
        # method to implement their specific logic.
        return self.get_success_response()

    def get_success_response(self):
        return redirect(self.get_success_url())

    def get_success_url(self):
        return str(self.success_url)


# ================
# Order submission
# ================


class PaymentDetailsView(OrderPlacementMixin, generic.TemplateView):
    """
    For taking the details of payment and creating the order.

    This view class is used by two separate URLs: 'payment-details' and
    'preview'. The `preview` class attribute is used to distinguish which is
    being used. Chronologically, `payment-details` (preview=False) comes before
    `preview` (preview=True).

    If sensitive details are required (e.g. a bankcard), then the payment details
    view should submit to the preview URL and a custom implementation of
    `validate_payment_submission` should be provided.

    - If the form data is valid, then the preview template can be rendered with
      the payment-details forms re-rendered within a hidden div so they can be
      re-submitted when the 'place order' button is clicked. This avoids having
      to write sensitive data to disk anywhere during the process. This can be
      done by calling `render_preview`, passing in the extra template context
      vars.

    - If the form data is invalid, then the payment details templates needs to
      be re-rendered with the relevant error messages. This can be done by
      calling `render_payment_details`, passing in the form instances to pass
      to the templates.

    The class is deliberately split into fine-grained methods, responsible for
    only one thing.  This is to make it easier to subclass and override just
    one component of functionality.

    All projects will need to subclass and customise this class as no payment
    is taken by default.
    """

    template_name = "oscar/checkout/payment_details.html"
    template_name_preview = "oscar/checkout/preview.html"

    # These conditions are extended at runtime depending on whether we are in
    # 'preview' mode or not.
    pre_conditions = [
        "check_basket_is_not_empty",
        "check_basket_is_valid",
        "check_user_email_is_captured",
        "check_shipping_data_is_captured",
    ]

    # If preview=True, then we render a preview template that shows all order
    # details ready for submission.
    preview = False

    def get_pre_conditions(self, request):
        if self.preview:
            # The preview view needs to ensure payment information has been
            # correctly captured.
            return self.pre_conditions + ["check_payment_data_is_captured"]
        return super().get_pre_conditions(request)

    def get_skip_conditions(self, request):
        if not self.preview:
            # Payment details should only be collected if necessary
            return ["skip_unless_payment_is_required"]
        return super().get_skip_conditions(request)

    def post(self, request, *args, **kwargs):
        # Posting to payment-details isn't the right thing to do.  Form
        # submissions should use the preview URL.
        if not self.preview:
            return http.HttpResponseBadRequest()

        # We use a custom parameter to indicate if this is an attempt to place
        # an order (normally from the preview page).  Without this, we assume a
        # payment form is being submitted from the payment details view. In
        # this case, the form needs validating and the order preview shown.
        if request.POST.get("action", "") == "place_order":
            return self.handle_place_order_submission(request)
        return self.handle_payment_details_submission(request)

    def handle_place_order_submission(self, request):
        """
        Handle a request to place an order.

        This method is normally called after the customer has clicked "place
        order" on the preview page. It's responsible for (re-)validating any
        form information then building the submission dict to pass to the
        `submit` method.

        If forms are submitted on your payment details view, you should
        override this method to ensure they are valid before extracting their
        data into the submission dict and passing it onto `submit`.
        """
        return self.submit(**self.build_submission())

    def handle_payment_details_submission(self, request):
        """
        Handle a request to submit payment details.

        This method will need to be overridden by projects that require forms
        to be submitted on the payment details view.  The new version of this
        method should validate the submitted form data and:

        - If the form data is valid, show the preview view with the forms
          re-rendered in the page
        - If the form data is invalid, show the payment details view with
          the form errors showing.

        """
        # No form data to validate by default, so we simply render the preview
        # page.  If validating form data and it's invalid, then call the
        # render_payment_details view.
        return self.render_preview(request)

    def render_preview(self, request, **kwargs):
        """
        Show a preview of the order.

        If sensitive data was submitted on the payment details page, you will
        need to pass it back to the view here so it can be stored in hidden
        form inputs.  This avoids ever writing the sensitive data to disk.
        """
        self.preview = True
        ctx = self.get_context_data(**kwargs)
        return self.render_to_response(ctx)

    def render_payment_details(self, request, **kwargs):
        """
        Show the payment details page

        This method is useful if the submission from the payment details view
        is invalid and needs to be re-rendered with form errors showing.
        """
        self.preview = False
        ctx = self.get_context_data(**kwargs)
        return self.render_to_response(ctx)

    def get_default_billing_address(self):
        """
        Return default billing address for user

        This is useful when the payment details view includes a billing address
        form - you can use this helper method to prepopulate the form.

        Note, this isn't used in core oscar as there is no billing address form
        by default.
        """
        if not self.request.user.is_authenticated:
            return None
        try:
            return self.request.user.addresses.get(is_default_for_billing=True)
        except UserAddress.DoesNotExist:
            return None

    def submit(
        self,
        user,
        basket,
        shipping_address,
        shipping_method,
        shipping_charge,
        billing_address,
        order_total,
        payment_kwargs=None,
        order_kwargs=None,
        surcharges=None,
    ):
        """
        Submit a basket for order placement.

        The process runs as follows:

         * Generate an order number
         * Freeze the basket so it cannot be modified any more (important when
           redirecting the user to another site for payment as it prevents the
           basket being manipulated during the payment process).
         * Attempt to take payment for the order
           - If payment is successful, place the order
           - If a redirect is required (e.g. PayPal, 3D Secure), redirect
           - If payment is unsuccessful, show an appropriate error message

        :basket: The basket to submit.
        :payment_kwargs: Additional kwargs to pass to the handle_payment
                         method. It normally makes sense to pass form
                         instances (rather than model instances) so that the
                         forms can be re-rendered correctly if payment fails.
        :order_kwargs: Additional kwargs to pass to the place_order method
        """
        if payment_kwargs is None:
            payment_kwargs = {}
        if order_kwargs is None:
            order_kwargs = {}

        # Taxes must be known at this point
        assert (
            basket.is_tax_known
        ), "Basket tax must be set before a user can place an order"
        assert (
            shipping_charge.is_tax_known
        ), "Shipping charge tax must be set before a user can place an order"

        # We generate the order number first as this will be used
        # in payment requests (ie before the order model has been
        # created).  We also save it in the session for multi-stage
        # checkouts (e.g. where we redirect to a 3rd party site and place
        # the order on a different request).
        order_number = self.generate_order_number(basket)
        self.checkout_session.set_order_number(order_number)
        logger.info(
            "Order #%s: beginning submission process for basket #%d",
            order_number,
            basket.id,
        )

        # Freeze the basket so it cannot be manipulated while the customer is
        # completing payment on a 3rd party site.  Also, store a reference to
        # the basket in the session so that we know which basket to thaw if we
        # get an unsuccessful payment response when redirecting to a 3rd party
        # site.
        self.freeze_basket(basket)
        self.checkout_session.set_submitted_basket(basket)

        # We define a general error message for when an unanticipated payment
        # error occurs.
        error_msg = _(
            "A problem occurred while processing payment for this "
            "order - no payment has been taken.  Please "
            "contact customer services if this problem persists"
        )

        signals.pre_payment.send_robust(sender=self, view=self)

        try:
            self.handle_payment(order_number, order_total, **payment_kwargs)
        except RedirectRequired as e:
            # Redirect required (e.g. PayPal, 3DS)
            logger.info("Order #%s: redirecting to %s", order_number, e.url)
            return http.HttpResponseRedirect(e.url)
        except UnableToTakePayment as e:
            # Something went wrong with payment but in an anticipated way.  Eg
            # their bankcard has expired, wrong card number - that kind of
            # thing. This type of exception is supposed to set a friendly error
            # message that makes sense to the customer.
            msg = str(e)
            logger.warning(
                "Order #%s: unable to take payment (%s) - restoring basket",
                order_number,
                msg,
            )
            self.restore_frozen_basket()

            # We assume that the details submitted on the payment details view
            # were invalid (e.g. expired bankcard).
            return self.render_payment_details(
                self.request, error=msg, **payment_kwargs
            )
        except PaymentError as e:
            # A general payment error - Something went wrong which wasn't
            # anticipated.  Eg, the payment gateway is down (it happens), your
            # credentials are wrong - that king of thing.
            # It makes sense to configure the checkout logger to
            # mail admins on an error as this issue warrants some further
            # investigation.
            msg = str(e)
            logger.error(
                "Order #%s: payment error (%s)", order_number, msg, exc_info=True
            )
            self.restore_frozen_basket()
            return self.render_preview(self.request, error=error_msg, **payment_kwargs)
        except Exception as e:
            # Unhandled exception - hopefully, you will only ever see this in
            # development...
            logger.exception(
                "Order #%s: unhandled exception while taking payment (%s)",
                order_number,
                e,
            )
            self.restore_frozen_basket()
            return self.render_preview(self.request, error=error_msg, **payment_kwargs)

        signals.post_payment.send_robust(sender=self, view=self)

        # If all is ok with payment, try and place order
        logger.info("Order #%s: payment successful, placing order", order_number)
        try:
            return self.handle_order_placement(
                order_number,
                user,
                basket,
                shipping_address,
                shipping_method,
                shipping_charge,
                billing_address,
                order_total,
                surcharges=surcharges,
                **order_kwargs,
            )
        except UnableToPlaceOrder as e:
            # It's possible that something will go wrong while trying to
            # actually place an order.  Not a good situation to be in as a
            # payment transaction may already have taken place, but needs
            # to be handled gracefully.
            msg = str(e)
            logger.error(
                "Order #%s: unable to place order - %s",
                order_number,
                msg,
                exc_info=True,
            )
            self.restore_frozen_basket()
            return self.render_preview(self.request, error=msg, **payment_kwargs)
        except Exception as e:
            # Hopefully you only ever reach this in development
            logger.exception(
                "Order #%s: unhandled exception while placing order (%s)",
                order_number,
                e,
            )
            error_msg = _(
                "A problem occurred while placing this order. Please contact customer services."
            )
            self.restore_frozen_basket()
            return self.render_preview(self.request, error=error_msg, **payment_kwargs)

    def get_template_names(self):
        return [self.template_name_preview] if self.preview else [self.template_name]


# =========
# Thank you
# =========


class ThankYouView(generic.DetailView):
    """
    Displays the 'thank you' page which summarises the order just submitted.
    """

    template_name = "oscar/checkout/thank_you.html"
    context_object_name = "order"

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        if self.object is None:
            return redirect(settings.OSCAR_HOMEPAGE)
        context = self.get_context_data(object=self.object)
        return self.render_to_response(context)

    def get_object(self, queryset=None):
        # We allow superusers to force an order thank-you page for testing
        order = None
        if self.request.user.is_superuser:
            kwargs = {}
            if "order_number" in self.request.GET:
                kwargs["number"] = self.request.GET["order_number"]
            elif "order_id" in self.request.GET:
                kwargs["id"] = self.request.GET["order_id"]
            if any(kwargs):
                order = Order._default_manager.filter(**kwargs).first()

        if not order:
            if "checkout_order_id" in self.request.session:
                order = Order._default_manager.filter(
                    pk=self.request.session["checkout_order_id"]
                ).first()
        return order

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        # Remember whether this view has been loaded.
        # Only send tracking information on the first load.
        if not ctx["order"].analytics_tracked:
            ctx["send_analytics_event"] = True
            ctx["order"].analytics_tracked = True
            ctx["order"].save(update_fields=["analytics_tracked"])
        else:
            ctx["send_analytics_event"] = False

        return ctx



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('order', '0002_auto_20141007_2032'),
    ]

    operations = [
        migrations.AlterField(
            model_name='order',
            name='date_placed',
            field=models.DateTimeField(db_index=True),
            preserve_default=True,
        ),
    ]



# Generated by Django 3.2.8 on 2021-10-27 11:42

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0025_attribute_code_uniquetogether_constraint'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='option',
            options={'ordering': ['order', 'name'], 'verbose_name': 'Option', 'verbose_name_plural': 'Options'},
        ),
        migrations.AddField(
            model_name='option',
            name='help_text',
            field=models.CharField(blank=True, help_text='Help text shown to the user on the add to basket form', max_length=255, null=True, verbose_name='Help text'),
        ),
        migrations.AddField(
            model_name='option',
            name='option_group',
            field=models.ForeignKey(blank=True, help_text='Select an option group if using type "Option" or "Multi Option"', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='product_options', to='catalogue.attributeoptiongroup', verbose_name='Option Group'),
        ),
        migrations.AddField(
            model_name='option',
            name='order',
            field=models.IntegerField(blank=True, db_index=True, help_text='Controls the ordering of product options on product detail pages', null=True, verbose_name='Ordering'),
        ),
        migrations.AlterField(
            model_name='option',
            name='type',
            field=models.CharField(choices=[('text', 'Text'), ('integer', 'Integer'), ('boolean', 'True / False'), ('float', 'Float'), ('date', 'Date'), ('select', 'Select'), ('radio', 'Radio'), ('multi_select', 'Multi select'), ('checkbox', 'Checkbox')], default='text', max_length=255, verbose_name='Type'),
        ),
    ]



from django.db import models
from django.urls import reverse
from django.utils.crypto import get_random_string
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy

from oscar.core.compat import AUTH_USER_MODEL


class AbstractWishList(models.Model):
    """
    Represents a user's wish lists of products.

    A user can have multiple wish lists, move products between them, etc.
    """

    # Only authenticated users can have wishlists
    owner = models.ForeignKey(
        AUTH_USER_MODEL,
        related_name="wishlists",
        on_delete=models.CASCADE,
        verbose_name=_("Owner"),
    )
    name = models.CharField(
        verbose_name=_("Name"), default=_("Default"), max_length=255
    )

    #: This key acts as primary key and is used instead of an int to make it
    #: harder to guess
    key = models.CharField(
        _("Key"), max_length=6, db_index=True, unique=True, editable=False
    )

    # Oscar core does not support public or shared wishlists at the moment, but
    # all the right hooks should be there
    PUBLIC, PRIVATE, SHARED = ("Public", "Private", "Shared")
    VISIBILITY_CHOICES = (
        (PRIVATE, _("Private - Only the owner can see the wish list")),
        (
            SHARED,
            _(
                "Shared - Only the owner and people with access to the"
                " obfuscated link can see the wish list"
            ),
        ),
        (PUBLIC, _("Public - Everybody can see the wish list")),
    )
    visibility = models.CharField(
        _("Visibility"), max_length=20, default=PRIVATE, choices=VISIBILITY_CHOICES
    )

    # Convention: A user can have multiple wish lists. The last created wish
    # list for a user shall be their "default" wish list.
    # If an UI element only allows adding to wish list without
    # specifying which one , one shall use the default one.
    # That is a rare enough case to handle it by convention instead of a
    # BooleanField.
    date_created = models.DateTimeField(
        _("Date created"), auto_now_add=True, editable=False, db_index=True
    )

    def __str__(self):
        return "%s's Wish List '%s'" % (self.owner, self.name)

    def save(self, *args, **kwargs):
        if not self.pk or kwargs.get("force_insert", False):
            self.key = self.__class__.random_key()
        super().save(*args, **kwargs)

    @classmethod
    def random_key(cls, length=6):
        """
        Get a unique random generated key
        """
        while True:
            key = get_random_string(
                length=length, allowed_chars="abcdefghijklmnopqrstuvwxyz0123456789"
            )
            if not cls._default_manager.filter(key=key).exists():
                return key

    def is_allowed_to_see(self, user):
        if user == self.owner:
            return True
        elif self.visibility == self.PUBLIC:
            return True
        elif self.visibility == self.SHARED and user.is_authenticated:
            return self.shared_emails.filter(email=user.email).exists()
        return False

    def is_allowed_to_edit(self, user):
        # currently only the owner can edit their wish list
        return user == self.owner

    class Meta:
        abstract = True
        app_label = "wishlists"
        ordering = (
            "owner",
            "date_created",
        )
        verbose_name = _("Wish List")

    def get_absolute_url(self):
        return reverse("customer:wishlists-detail", kwargs={"key": self.key})

    def add(self, product):
        """
        Add a product to this wishlist
        """
        lines = self.lines.filter(product=product)
        if len(lines) == 0:
            self.lines.create(product=product, title=product.get_title())
        else:
            line = lines[0]
            line.quantity += 1
            line.save()

    def get_shared_url(self):
        return reverse("wishlists:detail", kwargs={"key": self.key})

    @property
    def is_shareable(self):
        return self.visibility in [self.PUBLIC, self.SHARED]


class AbstractLine(models.Model):
    """
    One entry in a wish list. Similar to order lines or basket lines.
    """

    wishlist = models.ForeignKey(
        "wishlists.WishList",
        on_delete=models.CASCADE,
        related_name="lines",
        verbose_name=_("Wish List"),
    )
    product = models.ForeignKey(
        "catalogue.Product",
        verbose_name=_("Product"),
        related_name="wishlists_lines",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )
    quantity = models.PositiveIntegerField(_("Quantity"), default=1)
    #: Store the title in case product gets deleted
    title = models.CharField(pgettext_lazy("Product title", "Title"), max_length=255)

    def __str__(self):
        return "%sx %s on %s" % (self.quantity, self.title, self.wishlist.name)

    def get_title(self):
        if self.product:
            return self.product.get_title()
        else:
            return self.title

    class Meta:
        abstract = True
        app_label = "wishlists"
        # Enforce sorting by order of creation.
        ordering = ["pk"]
        unique_together = (("wishlist", "product"),)
        verbose_name = _("Wish list line")


class AbstractWishListSharedEmail(models.Model):
    """
    An email which is allowed to view and possibly edit the wishlist, if shared.
    The user will have to login/create an account with this email in order to view it.
    """

    wishlist = models.ForeignKey(
        "wishlists.WishList",
        on_delete=models.CASCADE,
        related_name="shared_emails",
        verbose_name=_("Wish List"),
    )
    email = models.EmailField(verbose_name=_("Email"))

    def __str__(self):
        return "%s - %s" % (self.wishlist.name, self.email)

    class Meta:
        abstract = True
        app_label = "wishlists"
        verbose_name = _("Wish list shared email")



# Generated by Django 2.0.7 on 2018-11-15 19:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0013_auto_20170821_1548'),
    ]

    operations = [
        migrations.AlterField(
            model_name='product',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Date created'),
        ),
        migrations.AlterField(
            model_name='productimage',
            name='display_order',
            field=models.PositiveIntegerField(db_index=True, default=0, help_text='An image with a display order of zero will be the primary image for a product', verbose_name='Display order'),
        ),
        migrations.AlterField(
            model_name='productrecommendation',
            name='ranking',
            field=models.PositiveSmallIntegerField(db_index=True, default=0, help_text='Determines order of the products. A product with a higher value will appear before one with a lower ranking.', verbose_name='Ranking'),
        ),
    ]



from django.contrib import messages


class FlashMessages(object):
    """
    Intermediate container for flash messages.

    This is useful as, at the time of creating the message, we don't know
    whether the response is an AJAX response or not.
    """

    def __init__(self):
        self.msgs = {}

    def add_message(self, level, message):
        self.msgs.setdefault(level, []).append(message)

    def add_messages(self, level, message_list):
        for msg in message_list:
            self.add_message(level, msg)

    def info(self, message):
        self.add_message(messages.INFO, message)

    def warning(self, message):
        self.add_message(messages.WARNING, message)

    def error(self, message):
        self.add_message(messages.ERROR, message)

    def success(self, message):
        self.add_message(messages.SUCCESS, message)

    def as_dict(self):
        payload = {}
        for level, msgs in self.msgs.items():
            tag = messages.DEFAULT_TAGS.get(level, "info")
            payload[tag] = [str(msg) for msg in msgs]
        return payload

    def apply_to_request(self, request):
        for level, msgs in self.msgs.items():
            for msg in msgs:
                messages.add_message(request, level, msg)



from django import forms
from django.utils.translation import gettext_lazy as _

from oscar.core.loading import get_class, get_model

ProductReview = get_model("reviews", "productreview")
DatePickerInput = get_class("oscar.forms.widgets", "DatePickerInput")


class DashboardProductReviewForm(forms.ModelForm):
    choices = (
        (ProductReview.APPROVED, _("Approved")),
        (ProductReview.REJECTED, _("Rejected")),
    )
    status = forms.ChoiceField(choices=choices, label=_("Status"))

    class Meta:
        model = ProductReview
        fields = ("title", "body", "score", "status")


class ProductReviewSearchForm(forms.Form):
    STATUS_CHOICES = (("", "------------"),) + ProductReview.STATUS_CHOICES
    keyword = forms.CharField(required=False, label=_("Keyword"))
    status = forms.ChoiceField(
        required=False, choices=STATUS_CHOICES, label=_("Status")
    )
    date_from = forms.DateTimeField(
        required=False, label=_("Date from"), widget=DatePickerInput
    )
    date_to = forms.DateTimeField(required=False, label=_("to"), widget=DatePickerInput)
    name = forms.CharField(required=False, label=_("Customer name"))

    def get_friendly_status(self):
        raw = int(self.cleaned_data["status"])
        for key, value in self.STATUS_CHOICES:
            if key == raw:
                return value
        return ""



from django.core import exceptions
from django.db import IntegrityError

from oscar.core.loading import get_model

Benefit = get_model("offer", "Benefit")
Condition = get_model("offer", "Condition")
Range = get_model("offer", "Range")


def _class_path(klass):
    return "%s.%s" % (klass.__module__, klass.__name__)


def create_range(range_class):
    """
    Create a custom range instance from the passed range class

    This function creates the appropriate database record for this custom
    range, including setting the class path for the custom proxy class.
    """
    if not hasattr(range_class, "name"):
        raise exceptions.ValidationError("A custom range must have a name attribute")

    # Ensure range name is text (not gettext wrapper)
    if range_class.name.__class__.__name__ == "__proxy__":
        raise exceptions.ValidationError(
            "Custom ranges must have text names (not gettext proxies)"
        )

    try:
        return Range.objects.create(
            name=range_class.name, proxy_class=_class_path(range_class)
        )
    except IntegrityError:
        raise ValueError("The passed range already exists in the database.")


def create_condition(condition_class, **kwargs):
    """
    Create a custom condition instance
    """
    return Condition.objects.create(proxy_class=_class_path(condition_class), **kwargs)


def create_benefit(benefit_class, **kwargs):
    """
    Create a custom benefit instance
    """
    # The custom benefit_class must override __str__ and description to
    # avoid a recursion error
    if benefit_class.description is Benefit.description:
        raise RuntimeError(
            "Your custom benefit must implement its own 'description' property"
        )
    return Benefit.objects.create(proxy_class=_class_path(benefit_class), **kwargs)



from collections.abc import Callable
from functools import wraps

from django.contrib.auth.decorators import user_passes_test
from django.core.exceptions import PermissionDenied
from django.shortcuts import render
from django.urls import reverse_lazy


def check_permissions(user, permissions):
    """
    Permissions can be a list or a tuple of lists. If it is a tuple,
    every permission list will be evaluated and the outcome will be checked
    for truthiness.
    Each item of the list(s) must be either a valid Django permission name
    (model.codename) or a property or method on the User model
    (e.g. 'is_active', 'is_superuser').

    Example usage:
    - permissions_required(['is_anonymous', ])
      would replace login_forbidden
    - permissions_required((['is_staff',], ['partner.dashboard_access']))
      allows both staff users and users with the above permission
    """

    def _check_one_permission_list(perms):
        regular_permissions = [perm for perm in perms if "." in perm]
        conditions = [perm for perm in perms if "." not in perm]
        # always check for is_active if not checking for is_anonymous
        if (
            conditions
            and "is_anonymous" not in conditions
            and "is_active" not in conditions
        ):
            conditions.append("is_active")
        attributes = [getattr(user, perm) for perm in conditions]
        # evaluates methods, explicitly casts properties to booleans
        passes_conditions = all(
            [
                attr() if isinstance(attr, Callable) else bool(attr)
                for attr in attributes
            ]
        )
        return passes_conditions and user.has_perms(regular_permissions)

    if not permissions:
        return True
    elif isinstance(permissions, list):
        return _check_one_permission_list(permissions)
    else:
        return any(_check_one_permission_list(perm) for perm in permissions)


def permissions_required(permissions, login_url=None):
    """
    Decorator that checks if a user has the given permissions.
    Accepts a list or tuple of lists of permissions (see check_permissions
    documentation).

    If the user is not logged in and the test fails, she is redirected to a
    login page. If the user is logged in, she gets a HTTP 403 Permission Denied
    message, analogous to Django's permission_required decorator.
    """
    if login_url is None:
        login_url = reverse_lazy("customer:login")

    def _check_permissions(user):
        outcome = check_permissions(user, permissions)
        if not outcome and user.is_authenticated:
            raise PermissionDenied
        else:
            return outcome

    return user_passes_test(_check_permissions, login_url=login_url)


def login_forbidden(view_func, template_name="oscar/login_forbidden.html", status=403):
    """
    Only allow anonymous users to access this view.
    """

    @wraps(view_func)
    def _checklogin(request, *args, **kwargs):
        if not request.user.is_authenticated:
            return view_func(request, *args, **kwargs)
        return render(request, template_name, status=status)

    return _checklogin
